import sys
import warnings
warnings.filterwarnings("ignore")
import click
import pandas as pd
from mgefinder import fastatools
from mgefinder import bowtie2tools
from mgefinder import sctools
from mgefinder import misc
from mgefinder import pysamtools
from mgefinder.inferseq import InferSequence, AlignedPairs
import pygogo as gogo
import pysam
from Bio import SeqIO
from collections import OrderedDict
from os.path import dirname, join
from random import randint
from snakemake import shell
from collections import defaultdict

verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger


import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


def _inferseq_assembly(pairsfile, bamfile, inferseq_assembly, inferseq_reference, min_perc_identity,
                       max_internal_softclip_prop, max_inferseq_size, min_inferseq_size, keep_intermediate, output_file):

    pairs = pd.read_csv(pairsfile, sep='\t', keep_default_na=False, na_values=[
        '-1.#IND', '1.#QNAN', '1.#IND', '-1.#QNAN', '#N/A', 'N/A', '#NA', 'NULL', 'NaN', '-NaN', 'nan', '-nan'])

    handle_empty_pairsfile(pairs, output_file)

    index_genome(inferseq_assembly)
    tmp_dir = dirname(output_file)

    context_width = 25
    context_inferer = InferSequenceContext(
        pairs, inferseq_assembly, bamfile, inferseq_reference, min_perc_identity,
        max_internal_softclip_prop, max_inferseq_size,
        min_inferseq_size, keep_intermediate,
        context_width, 'inferred_assembly_with_context', tmp_dir
    )

    inferred_sequences_with_context = context_inferer.infer_sequences()

    nocontext_inferer = InferSequence(
        pairs, inferseq_assembly, min_perc_identity, max_internal_softclip_prop, max_inferseq_size,
        min_inferseq_size, keep_intermediate, 'inferred_assembly_without_context', tmp_dir
    )

    inferred_sequences_without_context = nocontext_inferer.infer_sequences()

    inferred_sequences = pd.concat([inferred_sequences_with_context, inferred_sequences_without_context]).sort_values(
        ['pair_id', 'loc']
    )

    logger.info("Inferred sequences for %d pairs..." % len(set(list(inferred_sequences['pair_id']))))
    logger.info("Writing results to file %s..." % output_file)
    if pairs.shape[0] > 0:
        sample_id = list(pairs['sample'])[0]
        inferred_sequences.insert(0, 'sample', sample_id)
    else:
        inferred_sequences.insert(0, 'sample', None)

    inferred_sequences.to_csv(output_file, sep='\t', index=False)


class InferSequenceContext(InferSequence):

    context_width = None
    ref_bam = None
    ref_genome_dict = None

    def __init__(self, pairs, genome_fasta, ref_bam, ref_genome_fasta, min_perc_identity, max_internal_softclip_prop,
                 max_inferseq_size, min_inferseq_size, keep_intermediate, context_width, method_name='inferred_sequence',
                 tmp_dir='/tmp'):

        InferSequence.__init__(self, pairs, genome_fasta, min_perc_identity, max_internal_softclip_prop,
                               max_inferseq_size, min_inferseq_size, keep_intermediate, method_name, tmp_dir)

        self.ref_bam = pysam.AlignmentFile(ref_bam, 'rb')
        self.context_width = context_width
        self.ref_genome_dict = {rec.id: rec.seq for rec in SeqIO.parse(ref_genome_fasta, 'fasta')}

        self.all_aligned_pairs = defaultdict(lambda: AlignedPairsContext(self.context_width))


    def get_termini(self):

        context_termini = []
        for index, p in self.pairs.iterrows():
            pair_id, contig_name, pos_5p, pos_3p, seq_5p, seq_3p = p['pair_id'], p['contig'], p['pos_5p'], p['pos_3p'], \
                                                                   p['seq_5p'], p['seq_3p']
            seq_5p, seq_3p = seq_5p.rstrip('N'), seq_3p.lstrip('N')
            context_5p = self.get_sequence_context(self.ref_genome_dict[contig_name], contig_name,
                                                   pos_5p - self.context_width, pos_5p)
            context_3p = self.get_sequence_context(self.ref_genome_dict[contig_name], contig_name,
                                                   pos_3p + 1, pos_3p + 1 + self.context_width)

            context_termini.append({'pair_id': str(pair_id),
                                   'seq_5p': context_5p + seq_5p,
                                   'seq_3p': seq_3p + context_3p})

        return context_termini


    def get_sequence_context(self, contig, contig_name, start, end):

        add_start_n = 0
        add_end_n = 0
        if start < 0:
            add_start_n = abs(start)
            start = 0
        if end > len(contig):
            add_end_n = end - len(contig)
            end = len(contig)

        reference_sequence = 'N' * add_start_n + contig[start:end] + 'N' * add_end_n
        sequence_context_dict = initialize_sequence_context(reference_sequence, start, end)

        for read in self.ref_bam.fetch(contig_name, start, end):

            ref_positions = read.get_reference_positions(full_length=True)
            read_query = read.query_sequence
            read_qualities = read.query_qualities

            i = 0
            for pos in ref_positions:
                if pos is not None and pos in sequence_context_dict:
                    sequence_context_dict[pos][read_query[i]] += read_qualities[i]
                i += 1

        consensus = self.get_consensus_context(sequence_context_dict)
        return consensus


    def get_consensus_context(self, sequence_context_dict):
        start, end = min(sequence_context_dict.keys()), max(sequence_context_dict.keys()) + 1
        consensus = ''
        for pos in range(start, end):
            best_qual = 0
            best_nuc = ''
            for nuc in sequence_context_dict[pos]:
                qual = sequence_context_dict[pos][nuc]
                if qual > best_qual:
                    best_qual = qual
                    best_nuc = nuc
            consensus += best_nuc
        return consensus


    def get_inferred_sequence(self, forward_read, reverse_read, is_reverse):
        contig = forward_read.reference_name
        start = forward_read.reference_start
        end = reverse_read.reference_end

        inferred_sequence = ''.join(self.genome_dict[contig][start:end])

        inferred_sequence = sctools.left_softclipped_sequence_strict(forward_read) + \
                            inferred_sequence + \
                            sctools.right_softclipped_sequence_strict(reverse_read)

        inferred_sequence = inferred_sequence[self.context_width:-self.context_width]

        if is_reverse:
            inferred_sequence = misc.revcomp(inferred_sequence)

        contig_edge = False
        if sctools.is_left_softclipped_strict(forward_read) and \
                        sctools.left_softclipped_position(forward_read) < 0:
            contig_edge = True
        elif sctools.is_right_softclipped_strict(reverse_read) and \
                        sctools.right_softclipped_position(reverse_read) >= len(self.genome_dict[contig]):
            contig_edge = True


        return inferred_sequence, contig_edge

    def make_dataframe(self):

        outdict = OrderedDict([("pair_id", []), ("method", []), ("loc", []),
                               ("inferred_seq_length", []), ("inferred_seq", [])])

        for aligned_pairs in self.all_aligned_pairs:
            for pair in self.all_aligned_pairs[aligned_pairs].pairs:

                inferred_seq, contig_edge = self.get_inferred_sequence(
                    pair.forward_read, pair.reverse_read, pair.is_reverse()
                )
                outdict['pair_id'].append(int(pair.get_pair_id()))

                if contig_edge == True:
                    outdict['method'].append('inferred_assembly_with_half_context')
                else:
                    outdict['method'].append('inferred_assembly_with_full_context')

                outdict['loc'].append(pair.get_location())
                outdict['inferred_seq_length'].append(len(inferred_seq))
                outdict['inferred_seq'].append(inferred_seq)

        outdf = pd.DataFrame.from_dict(outdict).sort_values(['pair_id', 'loc']).reset_index(drop=True)
        return outdf


class AlignedPairsContext(AlignedPairs):

    context_width = None


    def __init__(self, context_width):
        AlignedPairs.__init__(self)
        self.context_width = context_width


    def add_read(self, read):

        if read.is_reverse:
            if read.query_name.split('_')[-1] == '1':
                self.reverse_reads_mate1.append(read)
                self.reverse_reads_mate1_positions[read.reference_name][read.reference_end-self.context_width] = read
            else:
                self.reverse_reads_mate2.append(read)
                self.reverse_reads_mate2_positions[read.reference_name][read.reference_end-self.context_width] = read
        else:
            if read.query_name.split('_')[-1] == '1':
                self.forward_reads_mate1.append(read)
                self.forward_reads_mate1_positions[read.reference_name][read.reference_start+self.context_width] = read
            else:
                self.forward_reads_mate2.append(read)
                self.forward_reads_mate2_positions[read.reference_name][read.reference_start+self.context_width] = read


    def filter_pairs_max_internal_softclip_prop(self, max_internal_softclip_prop):
        keep_pairs = list()
        for p in self.pairs:

            if sctools.is_left_softclipped_strict(p.forward_read) and \
                sctools.get_left_softclip_length(p.forward_read) > 1 and \
                sctools.is_right_softclipped_strict(p.reverse_read) and \
                sctools.get_right_softclip_length(p.reverse_read) > 1:
                continue

            if sctools.is_right_softclipped_strict(p.forward_read) and \
                p.forward_read.reference_end < p.reverse_read.reference_end and \
                sctools.right_softclip_proportion(p.forward_read) > max_internal_softclip_prop:
                continue

            if sctools.is_left_softclipped_strict(p.reverse_read) and \
                p.reverse_read.reference_start > p.forward_read.reference_start and \
                sctools.left_softclip_proportion(p.reverse_read) > max_internal_softclip_prop:
                continue

            keep_pairs.append(p)

        self.pairs = keep_pairs


    def get_closest_reverse_read(self, read):

        closest_reverse_read = None
        if read.query_name.split('_')[-1] == '1':
            positions = sorted(list(self.reverse_reads_mate2_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestLarger(positions, read.reference_start+self.context_width)
            if closest_pos is not None:
                closest_reverse_read = self.reverse_reads_mate2_positions[read.reference_name][closest_pos]
        else:
            positions = sorted(list(self.reverse_reads_mate1_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestLarger(positions, read.reference_start+self.context_width)
            if closest_pos is not None:
                closest_reverse_read = self.reverse_reads_mate1_positions[read.reference_name][closest_pos]

        print(read.query_name)
        return closest_reverse_read


    def get_closest_forward_read(self, read):

        closest_forward_read = None
        if read.query_name.split('_')[-1] == '1':
            positions = sorted(list(self.forward_reads_mate2_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestSmaller(positions, read.reference_end-self.context_width)
            if closest_pos is not None:
                closest_forward_read = self.forward_reads_mate2_positions[read.reference_name][closest_pos]
        else:
            positions = sorted(list(self.forward_reads_mate1_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestSmaller(positions, read.reference_end-self.context_width)
            if closest_pos is not None:
                closest_forward_read = self.forward_reads_mate1_positions[read.reference_name][closest_pos]

        return closest_forward_read



def get_inferred_sequences(pairs, genome_dict, add_softclipped_bases=False):

    inferred_sequences = []
    for read1, read2 in pairs:
        if read1.query_name.count('_') == 2:
            context_width = int(read1.query_name.split('_')[-2])
            name = read1.reference_name + ':' + str(read1.reference_start+context_width) + '-' + str(read2.reference_end-context_width)

            inferred_sequence = genome_dict[read1.reference_name][read1.reference_start:read2.reference_end]

            if add_softclipped_bases:
                inferred_sequence = sctools.left_softclipped_sequence_strict(read1) + inferred_sequence + sctools.right_softclipped_sequence_strict(read2)

            inferred_sequence = inferred_sequence[context_width:-context_width]

            if read1.query_name.split('_')[-1] == '2':
                inferred_sequence = misc.revcomp(inferred_sequence)

            contig_edge = False
            if sctools.is_left_softclipped_strict(read1) and \
                sctools.left_softclipped_position(read1) < 0:
                contig_edge = True
            elif sctools.is_right_softclipped_strict(read2) and \
                sctools.right_softclipped_position(read2) >= len(genome_dict[read2.reference_name]):
                contig_edge = True

        else:
            name = read1.reference_name + ':' + str(read1.reference_start) + '-' + str(read2.reference_end)
            inferred_sequence = genome_dict[read1.reference_name][read1.reference_start:read2.reference_end]

            if add_softclipped_bases:
                inferred_sequence = sctools.left_softclipped_sequence_strict(read1) + inferred_sequence + sctools.right_softclipped_sequence_strict(read2)

            if read1.query_name.split('_')[-1] == '2':
                inferred_sequence = misc.revcomp(inferred_sequence)

            contig_edge = False
            if sctools.is_left_softclipped_strict(read1) and \
                            sctools.left_softclipped_position(read1) < 0:
                contig_edge = True
            elif sctools.is_right_softclipped_strict(read2) and \
                            sctools.right_softclipped_position(read2) >= len(genome_dict[read2.reference_name]):
                contig_edge = True

        inferred_sequences.append((name, len(inferred_sequence), contig_edge, inferred_sequence))

    return inferred_sequences


def initialize_sequence_context(target_region, expanded_start, expanded_end):
    target_region_reads = defaultdict(lambda: defaultdict(int))
    i = 0
    for pos in range(expanded_start, expanded_end):
        target_region_reads[pos][target_region[i]] += 1
        i += 1
    return target_region_reads


def index_genome(inferseq_assembly):
    if not bowtie2tools.genome_is_indexed(inferseq_assembly):
        logger.info("Indexing inferseq assembly...")
        bowtie2tools.index_genome(inferseq_assembly)
    logger.info("Genome has been indexed...")


def handle_empty_pairsfile(pairs, output_file):
    if pairs.shape[0] == 0:
        outfile = pd.DataFrame(columns=['pair_id', 'method', 'loc', 'inferred_seq_length', 'inferred_seq'])

        if not output_file:
            output_file = 'mgefinder.inferseq_assembly.tsv'

        outfile.to_csv(output_file, sep='\t', index=False)
        logger.info("Empty pairs file, exiting...")
        sys.exit()



import warnings
warnings.filterwarnings("ignore")
import sys
import click
from mgefinder import bwatools, samtools
from os.path import dirname, basename, join, isfile
import pysam
from snakemake import shell
import pygogo as gogo

verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger


def read_sam_pairs(samfile):
    while True:
        yield(next(samfile), next(samfile))

def format_for_mgefinder(in_sam, out_bam, read_format='rb', delete_in_sam=False):

    samfile = pysam.AlignmentFile(in_sam, read_format)
    outbam= pysam.AlignmentFile(out_bam, "wb", template=samfile)
    sam_pairs = read_sam_pairs(samfile)

    for p1, p2 in sam_pairs:

        p1_qual = p1.tostring(samfile).split('\t')[10]
        p2_qual = p2.tostring(samfile).split('\t')[10]

        if p1.query_name != p2.query_name:

            click.echo("Fatal Error: Mismatched reads in bwa_tools.format()")
            sys.exit()

        if not p1.is_unmapped and p2.is_unmapped:
            p1.set_tag('MT', p2.query_sequence)
            p1.set_tag('MQ', p2_qual)
            outbam.write(p1)
        elif p1.is_unmapped and not p2.is_unmapped:
            p2.set_tag('MT', p1.query_sequence)
            p2.set_tag('MQ', p1_qual)
            outbam.write(p2)
        elif not p1.is_unmapped and not p2.is_unmapped:
            p1.set_tag('MT', p2.query_sequence)
            p1.set_tag('MQ', p2_qual)
            p2.set_tag('MT', p1.query_sequence)
            p2.set_tag('MQ', p1_qual)
            outbam.write(p1)
            outbam.write(p2)

    samfile.close()
    outbam.close()
    if delete_in_sam:
        shell('rm {in_sam}'.format(in_sam=in_sam))

    if isfile(out_bam):
        return out_bam
    else:
        return None

def _formatbam(in_sam, out_bam, single_end, keep_tmp_files):

    logger.info("Removing secondary alignments...")
    tmp_cleaned_bam = join(dirname(out_bam), '.'.join(basename(out_bam).split('.')[:-1]) + '.cleaned.bam.tmp')
    sorted_query_name = samtools.remove_secondary_alignments(in_sam, tmp_cleaned_bam, delete_in_bam=False)
    if not sorted_query_name:
        logger.info("Fatal error: Failed to remove secondary alignments...")
        sys.exit()
    logger.info("Successfully removed secondary alignments...\n")

    if not single_end:
        logger.info("Formatting bam file for use by mgefinder...")
        tmp_formatted_bam = join(dirname(out_bam), '.'.join(basename(out_bam).split('.')[:-1]) + '.formatted.bam.tmp')
        reformatted = format_for_mgefinder(tmp_cleaned_bam, tmp_formatted_bam, delete_in_sam=not keep_tmp_files)
        if not reformatted:
            logger.error("Fatal error: SAM file reformatting failed.")
            sys.exit()
        logger.info("SAM file successfully reformatted...\n")
    else:
        tmp_formatted_bam = tmp_cleaned_bam

    logger.info("Sorting the BAM file by chromosomal location...")
    sorted_bam = samtools.sort_coordinate(tmp_formatted_bam, out_bam, delete_in_bam=not keep_tmp_files)
    if not sorted_bam:
        logger.error("Fatal error: Failed to sort the BAM file.")
        sys.exit()
    logger.info("BAM file successfully sorted...\n")

    logger.info("Index the sorted BAM file...")
    indexed = samtools.index(out_bam)
    if not indexed:
        logger.info("Fatal error: Failed to index sorted BAM file")
        sys.exit()
    logger.info("BAM file successfully indexed...\n")


if __name__ == '__main__':
    _formatbam()


import sys
import warnings
warnings.filterwarnings("ignore")
import pandas as pd
from mgefinder import fastatools
from mgefinder import bowtie2tools
from mgefinder import sctools
from mgefinder import misc
from mgefinder import pysamtools
import pygogo as gogo
import pysam
from Bio import SeqIO
from collections import OrderedDict
from os.path import dirname, join
from random import randint
from snakemake import shell
from collections import defaultdict

verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger


import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


def _inferseq_database(pairsfile, inferseq_database, min_perc_identity, max_internal_softclip_prop,
                       max_edge_distance, output_file, keep_intermediate):

    index_database(inferseq_database)

    database_dict = {rec.id: rec.seq for rec in SeqIO.parse(inferseq_database, 'fasta')}

    tmp_dir = dirname(output_file)

    pairs = pd.read_csv(pairsfile, sep='\t', keep_default_na=False, na_values=[
        '-1.#IND', '1.#QNAN', '1.#IND', '-1.#QNAN', '#N/A','N/A', '#NA', 'NULL', 'NaN', '-NaN', 'nan', '-nan'])

    handle_empty_pairsfile(pairs, output_file)

    logger.info("Aligning pairs to database...")
    assembly_termini_fasta_prefix = write_termini_to_align_to_database(pairs, tmp_dir)
    assembly_outbam = join(tmp_dir, 'mgefinder.inferseq_database.' + str(randint(0, 1e20)) + '.bam')
    bowtie2tools.align_fasta_to_genome(
        assembly_termini_fasta_prefix + '.fasta',
        inferseq_database, assembly_outbam, silence=True,
        additional_flags='--all --score-min G,1,5'
    )

    logger.info("Inferring sequences from pairs aligned to database...")
    sequences_inferred_database = infer_sequences_database(assembly_outbam, database_dict,
                                                           min_perc_identity, max_internal_softclip_prop, max_edge_distance)

    if not keep_intermediate:
        shell('rm {fasta_prefix}* {outbam}*'.format(fasta_prefix=assembly_termini_fasta_prefix, outbam=assembly_outbam))

    method1 = make_dataframe(sequences_inferred_database, method='inferred_database')

    all_inferred_results = method1.sort_values(
        by=['pair_id', 'method']
    )

    all_inferred_results.loc[:, 'pair_id'] = list(map(str, map(int, list(all_inferred_results['pair_id']))))
    all_inferred_results = all_inferred_results.query("inferred_seq_length > 0")

    logger.info("Writing results to file %s..." % output_file)

    if not output_file:
        output_file = 'mgefinder.inferseq_database.tsv'

    if pairs.shape[0] > 0:
        sample_id = list(pairs['sample'])[0]
        all_inferred_results.insert(0, 'sample', sample_id)
    else:
        all_inferred_results.insert(0, 'sample', None)

    all_inferred_results.to_csv(output_file, sep='\t', index=False)


def get_inferred_sequences(pairs, genome_dict, add_softclipped_bases=False):

    inferred_sequences = []
    for read1, read2 in pairs:

        name = read1.reference_name + ':' + str(read1.reference_start) + '-' + str(read2.reference_end)
        inferred_sequence = genome_dict[read1.reference_name][read1.reference_start:read2.reference_end]

        if add_softclipped_bases:
            inferred_sequence = sctools.left_softclipped_sequence_strict(read1) + inferred_sequence + sctools.right_softclipped_sequence_strict(read2)

        if read1.is_read2:
            inferred_sequence = misc.revcomp(inferred_sequence)

        inferred_sequences.append((name, len(inferred_sequence), inferred_sequence))

    return inferred_sequences


def make_dataframe(inferred_sequences, method=None):
    outdict = OrderedDict([("pair_id", []), ("method", []), ("loc", []),
                           ("inferred_seq_length", []), ("inferred_seq", [])])
    for pair_id in inferred_sequences:
        for result in inferred_sequences[pair_id]:
            outdict['pair_id'].append(int(pair_id.split('_')[0]))
            outdict['method'].append(method)
            outdict['loc'].append(str(result[0]))
            outdict['inferred_seq_length'].append(result[1])
            outdict['inferred_seq'].append(''.join(result[2]))

    outdf = pd.DataFrame.from_dict(outdict)

    return outdf


def infer_sequences_database(bam_file, database_dict, min_perc_identity, max_internal_softclip_prop, max_edge_distance):
    bam = pysam.AlignmentFile(bam_file, 'rb')

    keep_reads = prefilter_reads(bam, database_dict, min_perc_identity, max_internal_softclip_prop, max_edge_distance)
    keep_pairs = get_pairs(keep_reads, database_dict, max_edge_distance)

    print("TOTAL PAIRS AFTER FILTER 1: %d" % count_total_pairs(keep_pairs))

    for pair_id in keep_pairs:
        keep_pairs[pair_id] = keep_best_alignment_score(keep_pairs[pair_id])
    print("TOTAL PAIRS AFTER FILTER 2: %d" % count_total_pairs(keep_pairs))

    inferred_sequences = defaultdict(list)
    for pair in keep_pairs:
        inferred_sequences[pair] = get_inferred_sequences(keep_pairs[pair], database_dict, add_softclipped_bases=True)

    return inferred_sequences


def count_total_pairs(pairs):
    count = 0
    for pair in pairs:
        count += len(pairs[pair])
    return count


def prefilter_reads(bam, database_dict, min_perc_identity, max_internal_softclip_prop, max_edge_distance):
    keep_reads = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    for read in bam:


        if pysamtools.get_perc_identity(read) < min_perc_identity:
            continue

        if not read.is_reverse:

            if sctools.is_right_softclipped_strict(read) and \
                sctools.right_softclipped_position(read) < len(database_dict[read.reference_name]) and \
                sctools.right_softclip_proportion(read) > max_internal_softclip_prop:
                continue

            elif read.reference_start > max_edge_distance:
                continue

            elif sctools.is_left_softclipped_strict(read) and \
                abs(0 - sctools.left_softclip_reference_start(read)) > max_edge_distance:
                continue

        if read.is_reverse:

            if sctools.is_left_softclipped_strict(read) and \
                sctools.left_softclipped_position(read) >= 0 and \
                sctools.left_softclip_proportion(read) > max_internal_softclip_prop:
                continue

            elif (len(database_dict[read.reference_name]) - read.reference_end) > max_edge_distance:
                continue

            elif sctools.is_right_softclipped_strict(read) and \
                abs(0 - (len(database_dict[read.reference_name]) - sctools.right_softclip_reference_end(read))) > max_edge_distance:
                continue

        pair_id, terminus_id = read.query_name.split('_')

        keep_reads[pair_id][read.reference_name][terminus_id].append(read)

    return keep_reads


def get_pairs(reads, database_dict, length_difference_max):

    keep_reads = defaultdict(list)
    for pair in reads:

        for ref in reads[pair]:

            if len(reads[pair][ref]['1']) > 0 and \
                len(reads[pair][ref]['2']) > 0 and \
                reads_mapped_both_ends(reads[pair][ref]['1'], reads[pair][ref]['2'], database_dict, length_difference_max):

                pairs = match_pairs(reads[pair][ref]['1'], reads[pair][ref]['2'])


                for p in pairs:

                    if not pair_mapped_both_ends(p, database_dict, length_difference_max):
                        continue

                    if p[0].is_reverse and (not p[1].is_reverse):
                        keep_reads[pair].append((p[1], p[0]))

                    elif (not p[0].is_reverse) and p[1].is_reverse:
                        keep_reads[pair].append((p[0], p[1]))

    return keep_reads


def reads_mapped_both_ends(reads1, reads2, database_dict, length_difference_max):
    fiveprime_read1_count = 0
    fiveprime_read2_count = 0
    threeprime_read1_count = 0
    threeprime_read2_count = 0


    for read in reads1 + reads2:
        if read.is_reverse and (len(database_dict[read.reference_name]) - read.reference_end) <= length_difference_max:

            if read.query_name.split('_')[-1] == '1':
                threeprime_read1_count += 1
            else:
                threeprime_read2_count += 1
        elif not read.is_reverse and read.reference_start <= length_difference_max:
            if read.query_name.split('_')[-1] == '1':
                fiveprime_read1_count += 1
            else:
                fiveprime_read2_count += 1

    if (fiveprime_read1_count > 0 and threeprime_read2_count > 0) or \
        (threeprime_read1_count > 0 and fiveprime_read2_count > 0):
        return True
    else:
        return False

def pair_mapped_both_ends(pair, database_dict, length_difference_max):
    has_fiveprime_read1 = False
    has_fiveprime_read2 = False
    has_threeprime_read1 = False
    has_threeprime_read2 = False

    read1, read2 = pair


    if read1.is_reverse and (len(database_dict[read1.reference_name]) - read1.reference_end) <= length_difference_max:

        has_threeprime_read1 = True

    elif not read1.is_reverse and read1.reference_start <= length_difference_max:

        has_fiveprime_read1 = True


    if read2.is_reverse and (len(database_dict[read2.reference_name]) - read2.reference_end) <= length_difference_max:

        has_threeprime_read2 = True

    elif not read2.is_reverse and read2.reference_start <= length_difference_max:

        has_fiveprime_read2 = True


    if (has_fiveprime_read1 and has_threeprime_read2) or \
            (has_threeprime_read1 and has_fiveprime_read2):
        return True
    else:
        return False


def match_pairs(reads1, reads2):
    pairwise_dist = []
    for i in range(len(reads1)):
        read1 = reads1[i]
        if read1.is_reverse:
            read1_pos = read1.reference_end
        else:
            read1_pos = read1.reference_start
        for j in range(len(reads2)):
            read2 = reads2[j]

            if read2.is_reverse:
                read2_pos = read2.reference_end
            else:
                read2_pos = read1.reference_start

            pairwise_dist.append((i, j, abs(read1_pos - read2_pos)))

    pairwise_dist_sorted = sorted(pairwise_dist, key=lambda x: x[2], reverse=True)
    pairs = []

    paired1 = set()
    paired2 = set()

    for s in pairwise_dist_sorted:
        if s[0] not in paired1 and s[1] not in paired2:
            pairs.append((reads1[s[0]], reads2[s[1]]))
            paired1.add(s[0])
            paired2.add(s[1])

    return pairs


def keep_best_alignment_score(reads):
    keep_reads = []

    best_score = 0
    for read1, read2 in reads:
        read_combined_alignment_score = read1.get_tag('AS') + read2.get_tag('AS')
        if read_combined_alignment_score > best_score:
            best_score = read_combined_alignment_score

    for read1, read2 in reads:
        read_combined_alignment_score = read1.get_tag('AS') + read2.get_tag('AS')
        if read_combined_alignment_score == best_score:
            keep_reads.append((read1, read2))

    return keep_reads


def index_database(inferseq_database):

    if not bowtie2tools.genome_is_indexed(inferseq_database):
        logger.info("Indexing inferseq database...")
        bowtie2tools.index_genome(inferseq_database)
    logger.info("Database has been indexed...")


def write_termini_to_align_to_database(pairs, tmp_dir):
    fasta_prefix = join(tmp_dir, 'mgefinder.inferseq_database.' + str(randint(0, 1e20)))

    writetermini = get_termini(pairs)

    fastatools.write_termini_to_unpaired_fasta(writetermini, fasta_prefix)

    return fasta_prefix


def get_termini(pairs):
    termini = []
    for index, p in pairs.iterrows():
        pair_id, seq_5p, seq_3p = p['pair_id'], p['seq_5p'], p['seq_3p']
        seq_5p, seq_3p = seq_5p.upper().rstrip('N'), seq_3p.upper().lstrip('N')
        termini.append({'pair_id':str(pair_id),
                       'seq_5p': seq_5p,
                       'seq_3p': seq_3p})
    return termini


def handle_empty_pairsfile(pairs, output_file):
    if pairs.shape[0] == 0:
        outfile = pd.DataFrame(columns=['pair_id', 'method', 'loc', 'inferred_seq_length', 'inferred_seq'])

        if not output_file:
            output_file = 'mgefinder.inferseq_database.tsv'

        outfile.to_csv(output_file, sep='\t', index=False)
        logger.info("Empty pairs file, exiting...")
        sys.exit()


import warnings
warnings.filterwarnings("ignore")
import sys
from snakemake import shell
from mgefinder.fastatools import read_fasta
from pandas import Series

def run_einverted(fasta, gap=12, threshold=15, match=3, mismatch=-4, outfile='einverted.tmp.out', outseq='einverted.tmp.outseq'):
    command = 'einverted -sequence {fasta} -gap {gap} -threshold {threshold} -match {match} ' \
              '-mismatch {mismatch} -outfile {outfile} -outseq {outseq} -auto Y -warning N'.format(
        fasta=fasta, gap=gap, threshold=threshold, match=match, mismatch=mismatch, outfile=outfile, outseq=outseq
    )
    #print(command)
    shell(command)


def read_emboss_seq_results(outseq_path):

    pair = []
    for rec in read_fasta(outseq_path):
        d = list(map(int, rec.id.split('_')))
        out = Series({'ir_pos_5p': d[1], 'ir_pos_3p':d[2], 'seq':str(rec.seq)})
        pair.append(out)
        if len(pair) == 2:
            yield pair
            pair = []





import sys
import warnings
warnings.filterwarnings("ignore")
import click
import pandas as pd
import pygogo as gogo
from collections import OrderedDict
from collections import defaultdict
from Bio.pairwise2 import format_alignment
from Bio import pairwise2

verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger


import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


def _inferseq_overlap(pairsfile, min_overlap_score, min_overlap_perc_identity, min_inferseq_size, output_file):

    pairs = pd.read_csv(pairsfile, sep='\t', keep_default_na=False, na_values=[
        '-1.#IND', '1.#QNAN', '1.#IND', '-1.#QNAN', '#N/A','N/A', '#NA', 'NULL', 'NaN', '-NaN', 'nan', '-nan'])

    handle_empty_pairsfile(pairs, output_file)

    logger.info("Inferring sequences from overlap...")
    sequences_inferred_from_overlap = infer_sequences_overlap(pairs, min_overlap_score, min_overlap_perc_identity)
    method1 = make_dataframe(sequences_inferred_from_overlap, method='inferred_overlap')

    method1.loc[:, 'pair_id'] = list(map(str, map(int, list(method1['pair_id']))))
    method1 = method1.query("inferred_seq_length >= @min_inferseq_size")

    logger.info("Writing results to file %s..." % output_file)

    if pairs.shape[0] > 0:
        sample_id = list(pairs['sample'])[0]
        method1.insert(0, 'sample', sample_id)
    else:
        method1.insert(0, 'sample', None)
    method1.to_csv(output_file, sep='\t', index=False)


def make_dataframe(inferred_sequences, method=None):
    outdict = OrderedDict([("pair_id", []), ("method", []), ("loc", []),
                           ("inferred_seq_length", []), ("inferred_seq", [])])
    for pair_id in inferred_sequences:
        for result in inferred_sequences[pair_id]:
            outdict['pair_id'].append(pair_id)
            outdict['method'].append(method)
            outdict['loc'].append(str(result[0]))
            outdict['inferred_seq_length'].append(result[1])
            outdict['inferred_seq'].append(''.join(result[2]))

    outdf = pd.DataFrame.from_dict(outdict)

    return outdf


def infer_sequences_overlap(pairs, min_overlap_score, min_overlap_perc_identity):
    sequences = defaultdict(list)
    for index, row in pairs.iterrows():
        pair_id, seq_5p, seq_3p = row['pair_id'], row['seq_5p'], row['seq_3p']
        merged_assembly, seq_5p_ostart, seq_5p_oend, seq_3p_ostart, seq_3p_oend  = find_overlap(
            seq_5p, seq_3p, min_overlap_score, min_overlap_perc_identity)
        if merged_assembly is not None:
            loc = 'seq_5p:' + str(seq_5p_ostart) + '-' + str(seq_5p_oend) + ';' + 'seq_3p:' + str(seq_3p_ostart) + '-' + str(seq_3p_oend)
            sequences[pair_id].append((loc, len(merged_assembly), merged_assembly))

    return sequences

def find_overlap(seq_5p, seq_3p, min_overlap_score, min_overlap_perc_identity, verbose=False):
    align_info = get_best_sliding_alignment(seq_5p, seq_3p)
    best_score, best_score_mismatches, best_r_start, best_r_end, best_q_start, best_q_end = align_info
    align_length = best_r_end - best_r_start

    merged_assembly = None

    if best_score >= min_overlap_score and (1-best_score_mismatches / float(align_length)) > min_overlap_perc_identity:
        if verbose:
            click.echo("They merged!")
            alignments = pairwise2.align.localms(seq_5p, seq_3p, 1, -1, -5, -5)
            print(format_alignment(*alignments[0]))

        else:
            merged_assembly = merge_overlapping_sequences(seq_5p[:best_q_start], seq_5p[best_q_start:best_q_end],
                                                          seq_3p[best_r_start:best_r_end], seq_3p[best_r_end:])


    return merged_assembly, best_q_start, best_q_end, best_r_start, best_r_end


def get_best_sliding_alignment(query_read, ref_read):

    #print("Query Read:", query_read)
    #print("Reference Read:", ref_read)

    max_len = min(len(query_read), len(ref_read))
    max_len_query, max_len_ref = False, False

    if max_len == len(query_read):
        max_len_query = True
    else:
        max_len_ref = True

    best_score = -sys.maxsize
    best_score_mismatches = 0
    best_r_start = 0
    best_r_end = 0
    best_q_start = 0
    best_q_end = 0

    q_start = len(query_read) - 1
    q_end = len(query_read)
    r_start = 0
    r_end = 1

    reached_max = False

    while q_start != q_end:
        score = 0
        mismatches = 0
        for i in range(q_end - q_start):

            if query_read[q_start:q_end][i] == ref_read[r_start:r_end][i]:
                score += 1

            else:
                score -= 1
                mismatches += 1

        if score > best_score:
            best_score = score
            best_score_mismatches = mismatches
            best_r_start = r_start
            best_r_end = r_end
            best_q_start = q_start
            best_q_end = q_end

        if q_end - q_start == max_len:
            reached_max = True

        if not reached_max:
            q_start -= 1
            r_end += 1

        elif reached_max and max_len_query:
            if r_end == len(ref_read):
                r_start += 1
                q_end -= 1
            else:
                r_start += 1
                r_end += 1

        elif reached_max and max_len_ref:
            if q_start == 0:
                r_start += 1
                q_end -= 1
            else:
                q_start -= 1
                q_end -= 1

    return best_score, best_score_mismatches, best_r_start, best_r_end, best_q_start, best_q_end

def handle_empty_pairsfile(pairs, output_file):
    if pairs.shape[0] == 0:
        outfile = pd.DataFrame(columns=['pair_id', 'method', 'loc', 'inferred_seq_length', 'inferred_seq'])

        if not output_file:
            output_file = 'mgefinder.inferseq_overlap.tsv'

        outfile.to_csv(output_file, sep='\t', index=False)
        logger.info("Empty pairs file, exiting...")
        sys.exit()

def merge_overlapping_sequences(start_seq, overlap_seq1, overlap_seq2, end_seq):

    merged_length = len(start_seq) + len(overlap_seq1) + len(end_seq)
    merged_seq = start_seq

    for i in range(len(overlap_seq1)):
        if len(merged_seq) > merged_length/2:
            merged_seq += overlap_seq2[i]
        else:
            merged_seq += overlap_seq1[i]
    merged_seq += end_seq

    return(merged_seq)


if __name__ == '__main__':
    inferseq_overlap()


from os.path import dirname, join
from snakemake import snakemake


def _workflow(workdir, snakefile, config, cores, memory, unlock, rerun_incomplete, keep_going):

    snakemake(snakefile, configfile=config, config={'wd': workdir, 'memory': memory}, cores=cores, unlock=unlock,
              force_incomplete=rerun_incomplete, keepgoing=keep_going)


import sys
import pysam
from collections import defaultdict, OrderedDict
from os.path import join
from random import randint
from mgefinder import fastatools, bowtie2tools, pysamtools, sctools, misc
import pygogo as gogo
from Bio import SeqIO
import pandas as pd
from snakemake import shell


verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger

class InferSequence:

    pairs = None
    genome_fasta = None
    method_name = None
    tmp_dir = None
    bam = None

    min_perc_identity = None
    max_internal_softclip_prop = None
    max_inferseq_size = None
    min_inferseq_size = None
    keep_intermediate = None

    all_aligned_pairs = None

    def __init__(self, pairs, genome_fasta, min_perc_identity, max_internal_softclip_prop,
                 max_inferseq_size, min_inferseq_size, keep_intermediate, method_name='inferred_sequence',
                 tmp_dir='/tmp'):

        self.pairs  = pairs
        self.genome_fasta = genome_fasta
        self.genome_dict = {rec.id: rec.seq for rec in SeqIO.parse(genome_fasta, 'fasta')}
        self.method_name = method_name
        self.tmp_dir = tmp_dir

        self.min_perc_identity = min_perc_identity
        self.max_internal_softclip_prop = max_internal_softclip_prop
        self.max_inferseq_size = max_inferseq_size
        self.min_inferseq_size = min_inferseq_size
        self.keep_intermediate = keep_intermediate

        self.all_aligned_pairs = defaultdict(AlignedPairs)

    def infer_sequences(self):

        self.__align_pairs_to_fasta()

        if not self.keep_intermediate:
            shell('rm {bam} {bam}.bai'.format(bam=self.bam.filename.decode('utf-8')))

        self.__prefilter_reads()

        for read in self.bam:
            name = read.query_name.split('_')[0]
            self.all_aligned_pairs[name].add_read(read)

        self.__match_forward_reverse_reads()

        self.__filter_smallest_overlapping()
        self.__filter_pairs_max_internal_softclip_prop()
        self.__filter_best_alignments()
        self.__filter_pairs_size()

        return self.make_dataframe()

    def __align_pairs_to_fasta(self):
        fasta_prefix = join(self.tmp_dir, 'mgefinder.inferseq.' + str(randint(0, 1e20)))

        writetermini = self.get_termini()

        fastatools.write_termini_to_unpaired_fasta(writetermini, fasta_prefix)

        logger.info("Aligning pairs to genome...")
        pair_bam_path = join(self.tmp_dir, 'mgefinder.inferseq.' + str(randint(0, 1e20)) + '.bam')
        bowtie2tools.align_fasta_to_genome(
            fasta_prefix + '.fasta',
            self.genome_fasta, pair_bam_path, silence=True,
            additional_flags='--all --score-min G,1,9'
        )
        self.bam = pysam.AlignmentFile(pair_bam_path, 'rb')


        if not self.keep_intermediate:
            shell('rm {fasta}'.format(fasta=fasta_prefix+'.fasta'))


    def get_termini(self):
        termini = []
        for index, p in self.pairs.iterrows():
            pair_id, seq_5p, seq_3p = p['pair_id'], p['seq_5p'], p['seq_3p']
            seq_5p, seq_3p = seq_5p.upper().rstrip('N'), seq_3p.upper().lstrip('N')
            termini.append({'pair_id': str(pair_id),
                           'seq_5p': seq_5p,
                           'seq_3p': seq_3p})
        return termini

    def __prefilter_reads(self):

        filtered_bam = []

        for read in self.bam:

            if pysamtools.get_perc_identity(read) < self.min_perc_identity:
                continue

            if not read.is_reverse:
                if sctools.is_left_softclipped_strict(read) and \
                    sctools.get_left_softclip_length(read) > 1 and \
                    sctools.left_softclipped_position(read) >= 0:
                    continue

            if read.is_reverse:
                if sctools.is_right_softclipped_strict(read) and \
                                sctools.get_right_softclip_length(read) > 1 and \
                                sctools.right_softclipped_position(read) < len(self.genome_dict[read.reference_name]):
                    continue

            filtered_bam.append(read)

        self.bam = filtered_bam

    def make_dataframe(self):

        outdict = OrderedDict([("pair_id", []), ("method", []), ("loc", []),
                               ("inferred_seq_length", []), ("inferred_seq", [])])

        for aligned_pairs in self.all_aligned_pairs:
            for pair in self.all_aligned_pairs[aligned_pairs].pairs:

                inferred_seq = self.get_inferred_sequence(
                    pair.forward_read, pair.reverse_read, pair.is_reverse()
                )
                outdict['pair_id'].append(int(pair.get_pair_id()))
                outdict['method'].append(self.method_name)
                outdict['loc'].append(pair.get_location())
                outdict['inferred_seq_length'].append(len(inferred_seq))
                outdict['inferred_seq'].append(inferred_seq)

        outdf = pd.DataFrame.from_dict(outdict).sort_values(['pair_id', 'method', 'loc']).reset_index(drop=True)
        return outdf

    def get_inferred_sequence(self, forward_read, reverse_read, is_reverse):
        contig, start, end = forward_read.reference_name, forward_read.reference_start, reverse_read.reference_end
        inferred_sequence = ''.join(self.genome_dict[contig][start:end])

        inferred_sequence = sctools.left_softclipped_sequence_strict(forward_read) + \
                            inferred_sequence + \
                            sctools.right_softclipped_sequence_strict(reverse_read)
        if is_reverse:
            inferred_sequence = misc.revcomp(inferred_sequence)

        return inferred_sequence

    def __match_forward_reverse_reads(self):
        for aligned_pairs in self.all_aligned_pairs:
            self.all_aligned_pairs[aligned_pairs].match_forward_reverse()

    def print_all_pairs(self):
        for aligned_pairs in self.all_aligned_pairs:
            self.all_aligned_pairs[aligned_pairs].print_pairs()

    def __filter_best_alignments(self):
        for aligned_pairs in self.all_aligned_pairs:
            self.all_aligned_pairs[aligned_pairs].filter_best_alignments()

    def __filter_smallest_overlapping(self):
        for aligned_pairs in self.all_aligned_pairs:
            self.all_aligned_pairs[aligned_pairs].filter_smallest_overlapping()

    def __filter_pairs_max_internal_softclip_prop(self):
        for aligned_pairs in self.all_aligned_pairs:
            self.all_aligned_pairs[aligned_pairs].filter_pairs_max_internal_softclip_prop(self.max_internal_softclip_prop)

    def __filter_pairs_size(self):
        for aligned_pairs in self.all_aligned_pairs:
            self.all_aligned_pairs[aligned_pairs].filter_pairs_size(self.max_inferseq_size, self.min_inferseq_size)


class AlignedPairs:


    forward_reads_mate1 = None
    forward_reads_mate2 = None
    forward_reads_mate1_positions = None
    forward_reads_mate2_positions = None

    reverse_reads_mate1 = None
    reverse_reads_mate2 = None
    reverse_reads_mate1_positions = None
    reverse_reads_mate2_positions = None

    pairs = None

    def __init__(self):

        self.forward_reads_mate1 = []
        self.forward_reads_mate2 = []
        self.forward_reads_mate1_positions = defaultdict(dict)
        self.forward_reads_mate2_positions = defaultdict(dict)

        self.reverse_reads_mate1 = []
        self.reverse_reads_mate2 = []
        self.reverse_reads_mate1_positions = defaultdict(dict)
        self.reverse_reads_mate2_positions = defaultdict(dict)

        self.pairs = set()

    def add_read(self, read):

        if read.is_reverse:
            if read.query_name.split('_')[-1] == '1':
                self.reverse_reads_mate1.append(read)
                self.reverse_reads_mate1_positions[read.reference_name][read.reference_end] = read
            else:
                self.reverse_reads_mate2.append(read)
                self.reverse_reads_mate2_positions[read.reference_name][read.reference_end] = read
        else:
            if read.query_name.split('_')[-1] == '1':
                self.forward_reads_mate1.append(read)
                self.forward_reads_mate1_positions[read.reference_name][read.reference_start] = read
            else:
                self.forward_reads_mate2.append(read)
                self.forward_reads_mate2_positions[read.reference_name][read.reference_start] = read


    def match_forward_reverse(self):

        added = set()
        for read in self.forward_reads_mate1 + self.forward_reads_mate2:
            closest_reverse_read = self.get_closest_reverse_read(read)
            if closest_reverse_read is not None:

                name = (read.query_name, read.reference_name, read.reference_start, read.reference_end,
                        closest_reverse_read.query_name, closest_reverse_read.reference_name,
                        closest_reverse_read.reference_start, closest_reverse_read.reference_end)

                if name not in added:
                    self.pairs.add(AlignedPair(read, closest_reverse_read))
                    added.add(name)

        for read in self.reverse_reads_mate1 + self.reverse_reads_mate2:
            closest_forward_read = self.get_closest_forward_read(read)
            if closest_forward_read is not None:
                name = (closest_forward_read.query_name, closest_forward_read.reference_name,
                        closest_forward_read.reference_start, closest_forward_read.reference_end,
                        read.query_name, read.reference_name, read.reference_start, read.reference_end,)

                if name not in added:
                    self.pairs.add(AlignedPair(closest_forward_read, read))
                    added.add(name)

    def get_closest_reverse_read(self, read):

        closest_reverse_read = None
        if read.query_name.split('_')[-1] == '1':
            positions = sorted(list(self.reverse_reads_mate2_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestLarger(positions, read.reference_start)
            if closest_pos is not None:
                closest_reverse_read = self.reverse_reads_mate2_positions[read.reference_name][closest_pos]
        else:
            positions = sorted(list(self.reverse_reads_mate1_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestLarger(positions, read.reference_start)
            if closest_pos is not None:
                closest_reverse_read = self.reverse_reads_mate1_positions[read.reference_name][closest_pos]

        return closest_reverse_read


    def get_closest_forward_read(self, read):

        closest_forward_read = None
        if read.query_name.split('_')[-1] == '1':
            positions = sorted(list(self.forward_reads_mate2_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestSmaller(positions, read.reference_end)
            if closest_pos is not None:
                closest_forward_read = self.forward_reads_mate2_positions[read.reference_name][closest_pos]
        else:
            positions = sorted(list(self.forward_reads_mate1_positions[read.reference_name].keys()))
            closest_pos = misc.takeClosestSmaller(positions, read.reference_end)
            if closest_pos is not None:
                closest_forward_read = self.forward_reads_mate1_positions[read.reference_name][closest_pos]

        return closest_forward_read


    def has_both_mates(self, read1, read2):

        has_one = False
        has_two = False

        if read1.query_name.split('_')[-1] == '1' or read2.query_name.split('_')[-1] == '1':
            has_one = True
        if read1.query_name.split('_')[-1] == '2' or read2.query_name.split('_')[-1] == '2':
            has_two = True

        return has_one and has_two


    def filter_best_alignments(self):
        best_alignment_score = -sys.maxsize
        for p in self.pairs:
            align_score = p.forward_read.get_tag('AS') + p.reverse_read.get_tag('AS')
            if align_score > best_alignment_score:
                best_alignment_score = align_score

        keep_pairs = list()
        for p in self.pairs:
            align_score = p.forward_read.get_tag('AS') + p.reverse_read.get_tag('AS')
            if align_score == best_alignment_score:
                keep_pairs.append(p)

        self.pairs = keep_pairs


    def filter_smallest_overlapping(self):
        forward_sites = defaultdict(lambda: defaultdict(set))
        reverse_sites = defaultdict(lambda: defaultdict(set))
        for p in self.pairs:
            forward_sites[p.forward_read.query_name][p.forward_read.reference_start].add(p)
            reverse_sites[p.reverse_read.query_name][p.reverse_read.reference_end].add(p)

        keep_pairs = set()
        for p in self.pairs:

            forward_sites_pairs = forward_sites[p.forward_read.query_name][p.forward_read.reference_start]
            reverse_sites_pairs = reverse_sites[p.reverse_read.query_name][p.reverse_read.reference_end]

            if len(forward_sites_pairs) == 1 and len(reverse_sites_pairs) == 1:
                keep_pairs.add(p)

            elif len(forward_sites_pairs) > 1 and len(reverse_sites_pairs) == 1:
                smallest_window_p = self.get_smallest_window_pair(forward_sites_pairs)
                keep_pairs.add(smallest_window_p)

            elif len(forward_sites_pairs) == 1 and len(reverse_sites_pairs) > 1:
                smallest_window_p = self.get_smallest_window_pair(reverse_sites_pairs)
                keep_pairs.add(smallest_window_p)

            elif len(forward_sites_pairs) > 1 and len(reverse_sites_pairs) > 1:
                smallest_window_p = self.get_smallest_window_pair(forward_sites_pairs.union(reverse_sites_pairs))
                keep_pairs.add(smallest_window_p)
            else:
                print()
                print("MISSED SOMETHING - DEBUG")
                sys.exit()

        self.pairs = list(keep_pairs)

    def get_smallest_window_pair(self, pairs):
        smallest_pair = None
        minlength = sys.maxsize
        for p in pairs:
            if p.get_aligned_pair_length() < minlength:
                minlength = p.get_aligned_pair_length()
                smallest_pair = p
        return smallest_pair

    def filter_pairs_max_internal_softclip_prop(self, max_internal_softclip_prop):
        keep_pairs = list()
        for p in self.pairs:

            if sctools.is_right_softclipped_strict(p.forward_read) and \
                p.forward_read.reference_end < p.reverse_read.reference_end and \
                sctools.right_softclip_proportion(p.forward_read) > max_internal_softclip_prop:
                continue

            if sctools.is_left_softclipped_strict(p.reverse_read) and \
                p.reverse_read.reference_start > p.forward_read.reference_start and \
                sctools.left_softclip_proportion(p.reverse_read) > max_internal_softclip_prop:
                continue

            keep_pairs.append(p)

        self.pairs = keep_pairs

    def filter_pairs_size(self, max_inferseq_size, min_inferseq_size):
        keep_pairs = list()
        for p in self.pairs:

            if p.get_aligned_pair_length() < min_inferseq_size or p.get_aligned_pair_length() > max_inferseq_size:
                continue
            keep_pairs.append(p)

        self.pairs = keep_pairs

    def print_pairs(self):

        for p in self.pairs:
            print(p)

class AlignedPair:
    forward_read = None
    reverse_read = None


    def __init__(self, forward_read, reverse_read):
        self.forward_read = forward_read
        self.reverse_read = reverse_read

    def get_alignment_score(self):
        return self.forward_read.get_tag('AS') + self.reverse_read.get_tag('AS')

    def get_aligned_pair_length(self):
        return self.reverse_read.reference_end - self.forward_read.reference_start

    def get_pair_id(self):
        return self.forward_read.query_name.split('_')[0]

    def get_location(self):
        return '{contig}:{start}-{end}'.format(
            contig=self.forward_read.reference_name,
            start=self.forward_read.reference_start,
            end=self.reverse_read.reference_end
        )

    def get_inferred_sequence_contig(self):
        return self.forward_read.reference_name

    def get_inferred_sequence_start(self):
        return self.forward_read.reference_start

    def get_inferred_sequence_end(self):
        return self.reverse_read.reference_end

    def is_reverse(self):
        return self.forward_read.query_name.split('_')[-1] == '2'

    def __str__(self):
        out = '{fname} - {fcontig} {fstart} {fend}, {rname} - {rcontig} {rstart} {rend}, LENGTH={length}'.format(
            fname = self.forward_read.query_name, fcontig = self.forward_read.reference_name,
            fstart = self.forward_read.reference_start, fend = self.forward_read.reference_end,
            rname = self.reverse_read.query_name, rcontig = self.reverse_read.reference_name,
            rstart = self.reverse_read.reference_start, rend = self.reverse_read.reference_end,
            length = self.get_aligned_pair_length()
        )

        return out





import warnings
warnings.filterwarnings("ignore")
import sys

def left_softclip_length(read):
    return read.cigartuples[0][1]

def right_softclip_length(read):
    return read.cigartuples[-1][1]

def left_softclipped_position(read):
    return read.get_reference_positions()[0]-1

def right_softclipped_position(read):
    return read.get_reference_positions()[-1] + 1

def is_right_softclipped_lenient(read):
    if read.cigartuples[-1][0] == 4:
        return True
    elif read.get_tag('MD')[-1] == '0' and not read.get_tag('MD')[-2].isdigit():
        return True
    else:
        return False

def is_left_softclipped_lenient(read):
    if read.cigartuples[0][0] == 4:
        return True
    elif read.get_tag('MD')[0] == '0':
        return True
    else:
        return False

def is_left_softclipped_lenient_at_site(read, contig, pos):
    if not is_left_softclipped_lenient(read):
        return False
    if left_softclipped_site_lenient(read) == (contig, pos):
        return True
    else:
        return False

def is_right_softclipped_lenient_at_site(read, contig, pos):
    if not is_right_softclipped_lenient(read):
        return False
    if right_softclipped_site_lenient(read) == (contig, pos):
        return True
    else:
        return False

def is_softclipped_lenient_at_site(read, contig, pos):
    if is_right_softclipped_lenient_at_site(read, contig, pos):
        return True
    elif is_left_softclipped_lenient_at_site(read, contig, pos):
        return True
    else:
        return False

def is_left_softclipped_strict(read):
    if read.cigartuples[0][0] == 4:
        return True
    else:
        return False

def is_right_softclipped_strict(read):
    if read.cigartuples[-1][0] == 4:
        return True
    else:
        return False

def get_right_softclip_length(read):
    if is_right_softclipped_lenient(read):
        if read.cigartuples[-1][0] == 4:
            return read.cigartuples[-1][1]
        else:
            return 1
    else:
        return 0

def get_right_softclip_length_strict(read):
    if is_right_softclipped_strict(read):
        return read.cigartuples[-1][1]
    else:
        return 0

def get_left_softclip_length(read):
    if is_left_softclipped_lenient(read):
        if read.cigartuples[0][0] == 4:
            return read.cigartuples[0][1]
        else:
            return 1
    else:
        return 0

def get_left_softclip_length_strict(read):
    if is_left_softclipped_strict(read):
        return read.cigartuples[0][1]
    else:
        return 0

def right_softclipped_site_lenient(read):
    if read.cigartuples[-1][0] == 4:
        return read.reference_name, read.get_reference_positions()[-1] + 1
    elif read.query_sequence[-1] != read.get_reference_sequence()[-1]:
        return read.reference_name, read.get_reference_positions()[-1]

def left_softclipped_site_lenient(read):
    if read.cigartuples[0][0] == 4:
        return read.reference_name, read.get_reference_positions()[0] - 1
    elif read.query_sequence[0] != read.get_reference_sequence()[0]:
        return read.reference_name, read.get_reference_positions()[0]

def right_softclipped_sequence(read):
    if is_right_softclipped_lenient(read):
        if read.cigartuples[-1][0] == 4:
            return read.query_sequence[-read.cigartuples[-1][1]:]
        else:
            return read.query_sequence[-1]
    else:
        return ''

def right_softclipped_sequence_strict(read):
    if is_right_softclipped_strict(read):
        return read.query_sequence[-read.cigartuples[-1][1]:]
    else:
        return ''

def right_softclip_qualities(read):
    if is_right_softclipped_lenient(read):
        if read.cigartuples[-1][0] == 4:
            return list(read.query_qualities)[-read.cigartuples[-1][1]:]
        else:
            return [list(read.query_qualities)[-1]]
    else:
        return []

def left_softclipped_sequence(read):
    if is_left_softclipped_lenient(read):
        if read.cigartuples[0][0] == 4:
            return read.query_sequence[:read.cigartuples[0][1]]
        else:
            return read.query_sequence[0]
    else:
        return ''

def left_softclipped_sequence_strict(read):
    if is_left_softclipped_strict(read):
        return read.query_sequence[:read.cigartuples[0][1]]
    else:
        return ''

def left_softclip_qualities(read):
    if is_left_softclipped_lenient(read):
        if read.cigartuples[0][0] == 4:
            return list(read.query_qualities)[:read.cigartuples[0][1]]
        else:
            return [list(read.query_qualities)[0]]
    else:
        return []

def is_double_softclipped_lenient(read):

    if is_left_softclipped_lenient(read) and is_right_softclipped_lenient(read):
        return True
    else:
        return False

def read_meets_min_alignment_inner_length(read, min_alignment_inner_length):

    if not is_double_softclipped_lenient(read):
        return True

    right_length = len(right_softclipped_sequence(read))
    left_length = len(left_softclipped_sequence(read))

    total_length = len(read.query_sequence)
    aligned_length = total_length - right_length - left_length

    if aligned_length >= min_alignment_inner_length:
        return True
    else:
        return False

def right_softclip_proportion(read):
    return len(right_softclipped_sequence_strict(read)) / len(read.query_sequence)

def left_softclip_proportion(read):
    return len(left_softclipped_sequence_strict(read)) / len(read.query_sequence)

def left_softclip_reference_start(read):
    return read.reference_start - len(left_softclipped_sequence_strict(read))

def right_softclip_reference_end(read):
    return read.reference_end + len(right_softclipped_sequence_strict(read))




import click

class CustomHelp(click.Group):

    def __init__(self, *args, **kwargs):
        self.help_priorities = {}
        super(CustomHelp, self).__init__(*args, **kwargs)

    def get_help(self, ctx):
        self.list_commands = self.list_commands_for_help
        return super(CustomHelp, self).get_help(ctx)

    def list_commands_for_help(self, ctx):
        """reorder the list of commands when listing the help"""
        commands = super(CustomHelp, self).list_commands(ctx)
        return (c[1] for c in sorted(
            (self.help_priorities.get(command, 1), command)
            for command in commands))

    def command(self, *args, **kwargs):
        """Behaves the same as `click.Group.command()` except capture
        a priority for listing command names in help.
        """
        help_priority = kwargs.pop('help_priority', 1)
        help_priorities = self.help_priorities

        def decorator(f):
            cmd = super(CustomHelp, self).command(*args, **kwargs)(f)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator


import sys
import warnings
warnings.filterwarnings("ignore")
import pygogo as gogo
import pandas as pd
from mgefinder import fastatools, cdhittools
from Bio import SeqIO
from snakemake import shell
from collections import defaultdict
import click
from os.path import join, isfile
from os import makedirs, rename
from shutil import rmtree
from mgefinder.bowtie2tools import index_genome

verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger
from tqdm import tqdm


def _makedatabase(inferseqfiles, minimum_size, maximum_size, threads, memory, force, output_dir, prefix):

    logger.info("Parsing inferseq files")
    if len(inferseqfiles) == 1 and is_path_list(inferseqfiles[0]):
        inferseqfiles = [l.strip() for l in open(inferseqfiles[0], 'r')]

    logger.info("Combining the inferseq files...")
    inferseq = combine_inferseq_files(inferseqfiles, minimum_size, maximum_size)

    database_maker = DatabaseMaker(inferseq, threads, memory, output_dir)

    try:
        makedirs(output_dir)

    except FileExistsError:

        if force:
            logger.info('Deleting old database directory...')
            rmtree(output_dir)
            makedirs(output_dir)
        else:
            logger.info('Output directory already exists. Use --force to overwrite directory.')
            sys.exit()

    if inferseq.shape[0] == 0:
        logger.info("No termini found in the input file...")

        clustered_seqs = database_maker.get_header_dataframe()

    else:

        cluster_fna = database_maker.run_cluster_seqs()
        outfile = join(output_dir, prefix+'.fna')
        rename(cluster_fna, outfile)
        database_maker.remove_int_files()
        logger.info("Indexing database for use with bowtie2")
        index_genome(outfile)

def is_path_list(f):
    with open(f) as infile:
        for l in infile:

            if len(l.strip().split()) > 1:
                return False

            if not isfile(l.strip()):
                return False

    return True


class DatabaseMaker:

    inferseq = None
    outfile = None

    threads = None
    memory = None

    seq_fasta_path = None
    cluster_100p_path = None
    cluster_mappings = None
    final_cluster_mappings = None

    def __init__(self, inferseq, threads, memory, output_dir):

        self.inferseq = inferseq
        self.pair_id_list = list(self.inferseq['pair_id'])

        self.inferseq.fillna('N', inplace=True)
        self.outdir = output_dir

        self.threads = threads
        self.memory = memory

        self.seq_fasta_path = output_dir + '/mgefinder.inferseq.fna'
        self.cluster_100p_path = output_dir + '/mgefinder.inferseq.100p.fna'
        self.cluster_perc_ident_path = output_dir + '/mgefinder.inferseq.perc_ident.fna'


    def run_cluster_seqs(self):

        self.create_inferseq_fasta()
        self.cluster_100p()
        self.cluster_perc_identity()

        return self.cluster_perc_ident_path


    def make_cluster_mappings(self):
        self.cluster_mappings = self.map_seqs_to_clusters()
        self.cluster_shared_pairs()


    def create_inferseq_fasta(self):

        seqs = list(self.inferseq['inferred_seq'])

        fastatools.write_sequences_to_fasta(seqs, self.seq_fasta_path)


    def cluster_100p(self):

        logger.info("Clustering at 100% identity...")

        cdhittools.cluster_100p(self.seq_fasta_path, self.cluster_100p_path)


    def cluster_perc_identity(self):

        cdhittools.cluster_reciprocal_identity(self.cluster_100p_path, self.cluster_perc_ident_path,
                                               threads=self.threads, memory=self.memory,
                                               alignment_coverage=0.99, perc_identity=0.99)


    def map_seqs_to_clusters(self):

        starting_ids = [rec.id for rec in SeqIO.parse(self.seq_fasta_path, 'fasta')]
        cdhit_100p_clusters = cdhittools.parse_cdhit_output_100p(self.seq_fasta_path, self.cluster_100p_path)
        cdhit_perc_identity_clusters = cdhittools.parse_cdhit_output_perc_identity(self.cluster_perc_ident_path)


        final_cluster_mappings = []

        unclustered_count = 0
        for ident in starting_ids:
            if ident in cdhit_100p_clusters.keys() and cdhit_100p_clusters[ident] in cdhit_perc_identity_clusters:
                final_cluster_mappings.append((ident, 'PCluster' + cdhit_perc_identity_clusters[cdhit_100p_clusters[ident]]))
            else:
                final_cluster_mappings.append((ident, 'UCluster' + str(unclustered_count)))
                unclustered_count += 1

        return final_cluster_mappings


    def cluster_shared_pairs(self):

        G = nx.Graph()

        pair_clusters = defaultdict(set)
        for pair_index in self.cluster_mappings:
            pair = self.pair_id_list[int(pair_index[0])]
            pair_clusters[pair].add(pair_index[1])

        for pair in pair_clusters.keys():

            nodes = list(pair_clusters[pair])
            for node in nodes:
                G.add_node(node)

            if len(nodes) > 1:

                for i in range(len(nodes)-1):
                    n1, n2 = nodes[i], nodes[i+1]
                    G.add_edge(n1, n2)


        final_clusters = dict()
        comps = list(nx.connected_components(G))
        for i in range(len(comps)):
            comp = comps[i]

            for clust in comp:
                final_clusters[clust] = "FCluster" + str(i+1)

        self.final_cluster_mappings = final_clusters


    def make_final_dataframe(self):

        out_cluster = pd.DataFrame(self.inferseq)
        out_cluster['cluster_perc_ident'] = [clust[1] for clust in self.cluster_mappings]
        out_cluster['cluster_final'] = [self.final_cluster_mappings[clust[1]] for clust in self.cluster_mappings]
        out_cluster = out_cluster[self.get_header_list()]

        return out_cluster

    def get_header_list(self):
        header = ['pair_id', 'method', 'loc', 'inferred_seq_length', 'cluster_perc_ident', 'cluster_final',
                  'inferred_seq']

        return header


    def get_header_dataframe(self):
        terminus_pairs = pd.DataFrame(columns=self.get_header_list())

        return terminus_pairs

    def remove_int_files(self):

        shell("rm %s" % ' '.join([
            self.seq_fasta_path+'*', self.cluster_100p_path + '*', self.cluster_perc_ident_path + '*'
        ]))




def combine_inferseq_files(inferseq_files, minimum_size, maximum_size):
    logger.info('Loading file {num1}/{num2}: {f}'.format(num1=1, num2=len(inferseq_files), f=inferseq_files[0]))
    inferseq = pd.read_csv(inferseq_files[0], sep='\t')
    for i, f in enumerate(inferseq_files[1:]):
        logger.info('Loading file {num1}/{num2}: {f}'.format(num1=i+2, num2=len(inferseq_files), f=f))
        inferseq = inferseq.append(pd.read_csv(f, sep='\t').
                                   query('inferred_seq_length >= @minimum_size').
                                   query('inferred_seq_length <= @maximum_size'))
    return inferseq



import warnings
warnings.filterwarnings("ignore")
from mgefinder import sctools
from mgefinder.find import SoftclipParser, SoftclipSite
import pandas as pd
import pysam
import pygogo as gogo

verbose=True
logger = gogo.Gogo(__name__, verbose=verbose).logger


def _recall(pairsfile, bamfile, min_alignment_quality, min_alignment_inner_length, large_insertion_cutoff, output_file):
    pairs = pd.read_csv(pairsfile, sep='\t')
    bam = pysam.AlignmentFile(bamfile)

    recaller = Recaller(bam, pairs, min_alignment_quality, min_alignment_inner_length, large_insertion_cutoff)

    recaller.parse_clipped_and_unclipped_read_info()
    recall_out = recaller.make_dataframe()

    if output_file:
        logger.info("Saving results to file %s" % output_file)
        recall_out.to_csv(output_file, sep='\t', index=False)

    return recall_out


class Recaller(SoftclipParser):

    pairs_dataframe = None

    def __init__(self, bam, pairs_dataframe, min_alignment_quality, min_alignment_inner_length, large_insertion_cutoff):
        SoftclipParser.__init__(self, bam, min_alignment_quality=min_alignment_quality,
                                min_alignment_inner_length=min_alignment_inner_length,
                                large_insertion_cutoff=large_insertion_cutoff)

        self.pairs_dataframe = pairs_dataframe
        self.load_pairs()


    def load_pairs(self):
        for index, row in self.pairs_dataframe.iterrows():
            self.softclipped_sites[row['contig']][row['pos_5p']] = SoftclipSite()
            self.softclipped_sites[row['contig']][row['pos_3p']] = SoftclipSite()


    def parse_clipped_and_unclipped_read_info(self):
        if verbose:
            logger.info("Getting clipped and unclipped read information near softclipped sites...")
            pass

        for contig in self.softclipped_sites:

            for pos in self.softclipped_sites[contig]:

                reads_at_site = self.get_reads_at_site(contig, pos)

                softclipped_5p_reads, softclipped_3p_reads = self.get_clipped_read_info_at_site(contig, pos, reads_at_site)

                runthrough_reads, small_insertion_5p_reads, small_insertion_3p_reads, \
                large_insertion_5p_reads, large_insertion_3p_reads, deletion_reads = \
                    self.get_unclipped_read_info_at_site(contig, pos, reads_at_site)

                self.softclipped_sites[contig][pos].add_softclip_5p_reads(softclipped_5p_reads)
                self.softclipped_sites[contig][pos].add_softclip_3p_reads(softclipped_3p_reads)

                self.softclipped_sites[contig][pos].add_runthrough_reads(runthrough_reads)
                self.softclipped_sites[contig][pos].add_small_insertion_5p_reads(small_insertion_5p_reads)
                self.softclipped_sites[contig][pos].add_small_insertion_3p_reads(small_insertion_3p_reads)
                self.softclipped_sites[contig][pos].add_large_insertion_5p_reads(large_insertion_5p_reads)
                self.softclipped_sites[contig][pos].add_large_insertion_3p_reads(large_insertion_3p_reads)
                self.softclipped_sites[contig][pos].add_deletion_reads(deletion_reads)

                upstream_deletion_reads, downstream_deletion_reads = None, None

                if pos - 1 >= 0:
                    upstream_deletion_reads = self.get_unclipped_read_info_at_site(
                        contig, pos - 1, reads_at_site, deletions_only=True
                    )

                if pos + 1 < self.contig_lengths[contig]:
                    downstream_deletion_reads = self.get_unclipped_read_info_at_site(
                        contig, pos + 1, reads_at_site, deletions_only=True
                    )

                if upstream_deletion_reads:
                    self.softclipped_sites[contig][pos].add_upstream_deletion_reads(upstream_deletion_reads)

                if downstream_deletion_reads:
                    self.softclipped_sites[contig][pos].add_downstream_deletion_reads(downstream_deletion_reads)


    def get_clipped_read_info_at_site(self, contig, pos, reads):

        softclipped_5p_reads = set()
        softclipped_3p_reads = set()

        for read in reads:

            if not self.passes_read_filters(read):
                continue

            if sctools.is_left_softclipped_lenient_at_site(read, contig, pos):
                softclipped_3p_reads.add(read)

            if sctools.is_right_softclipped_lenient_at_site(read, contig, pos):
                softclipped_5p_reads.add(read)

        return softclipped_5p_reads, softclipped_3p_reads


    def make_dataframe(self):

        column_names = ['contig', 'pos', 'orient', 'softclip_count_5p', 'softclip_count_3p', 'runthrough_count',
                        'small_insertion_count_5p', 'small_insertion_count_3p',
                        'large_insertion_count_5p', 'large_insertion_count_3p', 'deletion_count',
                        'upstream_deletion_count', 'downstream_deletion_count', 'total_count']

        outdata= dict()
        for contig in self.softclipped_sites:
            sorted_positions = sorted(list(self.softclipped_sites[contig].keys()))

            for pos in sorted_positions:
                site = self.softclipped_sites[contig][pos]

                outdata[len(outdata)] = [
                    contig, pos, '5p', site.get_softclip_5p_count(), site.get_softclip_3p_count(),
                    site.get_runthrough_count(), site.get_small_insertion_5p_count(), site.get_small_insertion_3p_count(),
                    site.get_large_insertion_5p_count(), site.get_large_insertion_3p_count(),
                    site.get_deletion_count(), site.get_upstream_deletion_count(), site.get_downstream_deletion_count(),
                    site.get_total_count()
                ]

                outdata[len(outdata)] = [
                    contig, pos, '3p', site.get_softclip_5p_count(), site.get_softclip_3p_count(),
                    site.get_runthrough_count(), site.get_small_insertion_5p_count(), site.get_small_insertion_3p_count(),
                    site.get_large_insertion_5p_count(), site.get_large_insertion_3p_count(),
                    site.get_deletion_count(), site.get_upstream_deletion_count(), site.get_downstream_deletion_count(),
                    site.get_total_count()
                ]

        out_df = pd.DataFrame.from_dict(outdata, orient='index', columns=column_names)

        return out_df



import sys
from mgefinder import pysamtools, sctools, misc, flanktrie
import pysam
from collections import defaultdict
import pandas as pd
from jellyfish import levenshtein_distance
import itertools
from scipy.sparse.csgraph import connected_components
import numpy as np
from tqdm import tqdm

import pygogo as gogo
verbose = True
logger = gogo.Gogo(__name__, verbose=False).logger



def _find(bamfile, min_softclip_length, min_softclip_count, min_alignment_quality, min_alignment_inner_length,
                min_distance_to_mate, min_softclip_ratio, max_indel_ratio, large_insertion_cutoff,
                min_count_consensus, sample_id, output_file=None):

    bam = pysam.AlignmentFile(bamfile, 'rb')

    softclip_parser = SoftclipParser(
        bam, verbose=True,
        min_softclip_length=min_softclip_length,
        min_softclip_count=min_softclip_count,
        min_alignment_quality=min_alignment_quality,
        min_alignment_inner_length=min_alignment_inner_length,
        min_distance_to_mate=min_distance_to_mate,
        min_softclip_ratio=min_softclip_ratio,
        max_indel_ratio=max_indel_ratio,
        large_insertion_cutoff=large_insertion_cutoff,
        min_count_consensus=min_count_consensus
    )

    softclip_parser.parse_softclips()
    softclip_parser.filter_softclips_minlength()
    softclip_parser.filter_softclips_mincount()
    softclip_parser.filter_softclips_mindistance()
    softclip_parser.parse_unclipped_read_info()
    softclip_parser.filter_softclips_count_ratios()
    softclip_parser.filter_softclips_mindistance()
    softclip_parser.make_consensus_sequences()
    softclip_parser.filter_consensus_sequences_minlength()
    softclip_parser.filter_consensus_sequences_mincount()
    softclip_parser.filter_multiple_consensus_sequences()

    softclip_parser.filter_softclips_mindistance()

    final_df = softclip_parser.make_dataframe()
    final_df.insert(0, 'terminus_id', list(final_df.index+1))
    final_df.insert(0, 'sample', sample_id)

    if output_file:
        logger.info("Saving results to file %s" % output_file)
        final_df.to_csv(output_file, sep='\t', index=False)

    return final_df


class SoftclipParser:

    softclipped_sites = None
    bam = None
    contig_lengths = None
    verbose = None

    min_alignment_quality = None
    min_alignment_inner_length = None
    min_softclip_length = None
    min_softclip_count = None
    min_distance_to_mate = None
    min_softclip_ratio = None
    max_indel_ratio = None
    large_insertion_cutoff = None
    min_count_consensus = None


    def __init__(self, bam, verbose=True, min_alignment_quality=20, min_alignment_inner_length=21,
                 min_softclip_length=4, min_softclip_count=4, min_distance_to_mate=22,
                 min_softclip_ratio=0.15, max_indel_ratio=0.03, large_insertion_cutoff=30,
                 min_count_consensus=2):
        self.verbose = verbose
        self.bam = bam
        self.contig_lengths = pysamtools.get_bam_contig_dict(bam)

        self.softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))
        self.min_alignment_quality = min_alignment_quality
        self.min_alignment_inner_length = min_alignment_inner_length
        self.min_softclip_length = min_softclip_length
        self.min_softclip_count = min_softclip_count
        self.min_distance_to_mate = min_distance_to_mate
        self.min_softclip_ratio = min_softclip_ratio
        self.max_indel_ratio = max_indel_ratio
        self.large_insertion_cutoff = large_insertion_cutoff
        self.min_count_consensus = min_count_consensus


    def parse_softclips(self):

        if self.verbose:
            logger.info("Parsing softclipped sites from provided BAM file...")

        read_count = 0
        for read in self.bam:

            read_count += 1

            if self.verbose and read_count % 100000 == 0:
                logger.info("\tAfter checking %d reads, %d softclipped sites found..." % (read_count, self.count_softclips()))
                pass

            if not self.passes_read_filters(read):
                continue

            if sctools.is_right_softclipped_lenient(read):
                self.parse_right_softclipped_read(read)

            if sctools.is_left_softclipped_lenient(read):
                self.parse_left_softclipped_read(read)


    def filter_softclips_minlength(self):

        filtered_softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:

                if not self.softclipped_sites[contig][pos].meets_minlength_5p:
                    self.softclipped_sites[contig][pos].keep_softclips_5p = False

                if not self.softclipped_sites[contig][pos].meets_minlength_3p:
                    self.softclipped_sites[contig][pos].keep_softclips_3p = False

                if self.softclipped_sites[contig][pos].keep_softclips_5p or \
                    self.softclipped_sites[contig][pos].keep_softclips_3p:
                    filtered_softclipped_sites[contig][pos] = self.softclipped_sites[contig][pos]

        self.softclipped_sites = filtered_softclipped_sites

        if verbose:
            logger.info("After filtering by minimum softclip length of %d, %d sites remain" % (self.min_softclip_length, self.count_softclips()))
            pass


    def filter_softclips_mincount(self):
        filtered_softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                softclip_site = self.softclipped_sites[contig][pos]

                if softclip_site.get_softclip_5p_count() < self.min_softclip_count:
                    softclip_site.keep_softclips_5p = False

                if softclip_site.get_softclip_3p_count() < self.min_softclip_count:
                    softclip_site.keep_softclips_3p = False

                if softclip_site.keep_softclips_5p or softclip_site.keep_softclips_3p:
                    filtered_softclipped_sites[contig][pos] = softclip_site

        self.softclipped_sites = filtered_softclipped_sites

        if verbose:
            logger.info("After filtering by minimum softclipped read count of %d, %d sites remain" % (
                self.min_softclip_count, self.count_softclips()))
            pass

    def filter_softclips_mindistance(self):

        softclip_5p_positions = self.get_softclip_5p_positions()
        softclip_3p_positions = self.get_softclip_3p_positions()

        filtered_softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                softclip_site = self.softclipped_sites[contig][pos]

                if softclip_site.get_softclip_5p_count() > 0:

                    if not self.has_nearby_3p_mate(pos, softclip_3p_positions[contig]):
                        softclip_site.keep_softclips_5p = False

                if softclip_site.get_softclip_3p_count() > 0:

                    if not self.has_nearby_5p_mate(pos, softclip_5p_positions[contig]):
                        softclip_site.keep_softclips_3p = False

                if softclip_site.keep_softclips_5p or softclip_site.keep_softclips_3p:
                    filtered_softclipped_sites[contig][pos] = softclip_site

        self.softclipped_sites = filtered_softclipped_sites

        if verbose:
            logger.info("After filtering by minimum nearest mate distance %d, %d sites remain" % (
                self.min_distance_to_mate, self.count_softclips()))
            pass


    def filter_softclips_count_ratios(self):

        filtered_softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                softclip_site = self.softclipped_sites[contig][pos]

                if softclip_site.keep_softclips_5p:

                    if softclip_site.get_softclip_ratio_5p() < self.min_softclip_ratio or \
                        softclip_site.get_indel_ratio_5p() > self.max_indel_ratio:
                            softclip_site.keep_softclips_5p = False

                if softclip_site.keep_softclips_3p:

                    if softclip_site.get_softclip_ratio_3p() < self.min_softclip_ratio or \
                        softclip_site.get_indel_ratio_3p() > self.max_indel_ratio:
                        softclip_site.keep_softclips_3p = False

                if softclip_site.keep_softclips_5p or softclip_site.keep_softclips_3p:

                    if softclip_site.get_downstream_deletion_ratio() > self.max_indel_ratio or \
                       softclip_site.get_upstream_deletion_ratio() > self.max_indel_ratio:
                        softclip_site.keep_softclips_5p = False
                        softclip_site.keep_softclips_3p = False

                if softclip_site.keep_softclips_5p or softclip_site.keep_softclips_3p:
                    filtered_softclipped_sites[contig][pos] = softclip_site

        self.softclipped_sites = filtered_softclipped_sites

        if verbose:
            logger.info("After filtering by minimum softclip ratio of %f and a "
                        "maximum indel ratio of %f, %d sites remain" % (
                self.min_softclip_ratio, self.max_indel_ratio, self.count_softclips()))
            pass

    def filter_consensus_sequences_minlength(self):

        filtered_softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                softclip_site = self.softclipped_sites[contig][pos]


                if softclip_site.keep_softclips_5p:

                    consensus_seqs = []
                    for consensus in softclip_site.consensus_sequences_5p:

                        if len(consensus[1]) >= self.min_softclip_length:
                            consensus_seqs.append(consensus)

                    softclip_site.consensus_sequences_5p = consensus_seqs
                    if len(consensus_seqs) == 0:
                        softclip_site.keep_softclips_5p = False

                if softclip_site.keep_softclips_3p:

                    consensus_seqs = []
                    for consensus in softclip_site.consensus_sequences_3p:

                        if len(consensus[1]) >= self.min_softclip_length:
                            consensus_seqs.append(consensus)

                    softclip_site.consensus_sequences_3p = consensus_seqs
                    if len(consensus_seqs) == 0:
                        softclip_site.keep_softclips_3p = False

                if softclip_site.keep_softclips_5p or softclip_site.keep_softclips_3p:
                    filtered_softclipped_sites[contig][pos] = softclip_site


        self.softclipped_sites = filtered_softclipped_sites

        if verbose:
            logger.info("After filtering consensus sequences by a minimum length of %d, %d terminus sequences remain" % (
                            self.min_softclip_count, self.count_consensus_seqs())
                        )

    def filter_consensus_sequences_mincount(self):

        filtered_softclipped_sites = defaultdict(lambda: defaultdict(SoftclipSite))

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                softclip_site = self.softclipped_sites[contig][pos]

                if softclip_site.keep_softclips_5p:

                    consensus_seqs = []
                    for consensus in softclip_site.consensus_sequences_5p:

                        if consensus[0] >= self.min_softclip_count:
                            consensus_seqs.append(consensus)
                    softclip_site.consensus_sequences_5p = consensus_seqs
                    if len(consensus_seqs) == 0:
                        softclip_site.keep_softclips_5p = False

                if softclip_site.keep_softclips_3p:

                    consensus_seqs = []
                    for consensus in softclip_site.consensus_sequences_3p:

                        if consensus[0] >= self.min_softclip_count:
                            consensus_seqs.append(consensus)
                    softclip_site.consensus_sequences_3p = consensus_seqs
                    if len(consensus_seqs) == 0:
                        softclip_site.keep_softclips_3p = False

                if softclip_site.keep_softclips_5p or softclip_site.keep_softclips_3p:
                    filtered_softclipped_sites[contig][pos] = softclip_site

        self.softclipped_sites = filtered_softclipped_sites

        if verbose:
            logger.info("After filtering consensus sequences by a minimum softclip count of %d, %d terminus sequences remain" % (
                self.min_softclip_count, self.count_consensus_seqs())
                        )

    def filter_multiple_consensus_sequences(self):

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                softclip_site = self.softclipped_sites[contig][pos]

                if len(softclip_site.consensus_sequences_5p) > 1:
                    best_consensus = self.get_best_consensus_sequence(softclip_site.consensus_sequences_5p)
                    softclip_site.consensus_sequences_5p = best_consensus

                if len(softclip_site.consensus_sequences_3p) > 1:
                    best_consensus = self.get_best_consensus_sequence(softclip_site.consensus_sequences_3p)
                    softclip_site.consensus_sequences_3p = best_consensus

        if verbose:
            logger.info("After selecting the best consensus sequence at each site, %d terminus sequences remain." % (
                self.count_consensus_seqs()
            ))


    def get_best_consensus_sequence(self, consensus_seqs):
        return [list(sorted(consensus_seqs, key=lambda x: (x[0], len(x[1]))))[-1]]


    def parse_unclipped_read_info(self):
        if verbose:
            logger.info("Getting unclipped read information near softclipped sites...")
            pass

        pbar_contigs = tqdm(self.softclipped_sites, desc='Contigs', leave=False)

        for contig in pbar_contigs:

            pbar_contigs.set_description("Contig %s" % contig)

            pbar_positions = tqdm(self.softclipped_sites[contig], desc='Position', leave=False)
            for pos in pbar_positions:

                pbar_positions.set_description("Position %d" % pos)

                softclip_site = self.softclipped_sites[contig][pos]

                reads_at_site = self.get_reads_at_site(contig, pos)

                runthrough_reads, small_insertion_5p_reads, small_insertion_3p_reads, \
                large_insertion_5p_reads, large_insertion_3p_reads, deletion_reads = \
                    self.get_unclipped_read_info_at_site(contig, pos, reads_at_site)

                softclip_site.add_runthrough_reads(runthrough_reads)
                softclip_site.add_small_insertion_5p_reads(small_insertion_5p_reads)
                softclip_site.add_small_insertion_3p_reads(small_insertion_3p_reads)
                softclip_site.add_large_insertion_5p_reads(large_insertion_5p_reads)
                softclip_site.add_large_insertion_3p_reads(large_insertion_3p_reads)
                softclip_site.add_deletion_reads(deletion_reads)

                upstream_deletion_reads, downstream_deletion_reads = None, None

                if pos - 1 >= 0:
                    upstream_deletion_reads = self.get_unclipped_read_info_at_site(
                        contig, pos - 1, reads_at_site, deletions_only=True
                    )

                if pos + 1 < self.contig_lengths[contig]:
                    downstream_deletion_reads = self.get_unclipped_read_info_at_site(
                        contig, pos + 1, reads_at_site, deletions_only=True
                    )

                if upstream_deletion_reads:
                    softclip_site.add_upstream_deletion_reads(upstream_deletion_reads)

                if downstream_deletion_reads:
                    softclip_site.add_downstream_deletion_reads(downstream_deletion_reads)


    def get_reads_at_site(self, contig, pos):

        reads = []

        start = pos - 1
        end = pos + 2

        if start < 0:
            start = 0
        if end > self.contig_lengths[contig]:
            end = self.contig_lengths[contig]

        for read in self.bam.fetch(contig, start, end):
            if self.passes_read_filters(read):
                reads.append(read)

        return reads


    def get_unclipped_read_info_at_site(self, contig, pos, reads, deletions_only=False):

        runthrough_reads = set()
        small_insertion_5p_reads = set()
        small_insertion_3p_reads = set()
        large_insertion_5p_reads = set()
        large_insertion_3p_reads = set()
        deletion_reads = set()

        for read in reads:

            if sctools.is_softclipped_lenient_at_site(read, contig, pos):
                continue

            processed_read = self.process_aligned_blocks_at_site(pos, read, deletions_only)
            if processed_read is None:
                continue
            elif processed_read == 'runthrough':
                runthrough_reads.add(read.query_name)
            elif processed_read == 'small_insertion_5p':
                small_insertion_5p_reads.add(read.query_name)
            elif processed_read == 'small_insertion_3p':
                small_insertion_3p_reads.add(read.query_name)
            elif processed_read == 'large_insertion_5p':
                large_insertion_5p_reads.add(read.query_name)
            elif processed_read == 'large_insertion_3p':
                large_insertion_3p_reads.add(read.query_name)
            else:
                deletion_reads.add(read.query_name)

        if deletions_only:
            return deletion_reads
        else:
            return runthrough_reads, small_insertion_5p_reads, small_insertion_3p_reads, \
                   large_insertion_5p_reads, large_insertion_3p_reads, deletion_reads


    def process_aligned_blocks_at_site(self, position, read, deletions_only):

        blocks = read.get_blocks()

        runthrough = False
        small_insertion_5p = False
        small_insertion_3p = False
        large_insertion_5p = False
        large_insertion_3p = False
        deletion = False

        for b in blocks:

            if self.block_overlaps_site(b, position):
                runthrough = True

        if len(blocks) > 1:
            for i in range(len(blocks)-1):
                block1, block2 = blocks[i], blocks[i+1]
                if block1[1] == block2[0]:

                    if deletions_only:
                        continue

                    if position == block1[1]-1:
                        insertion_length = pysamtools.get_insertion_length(position, read)
                        if insertion_length >= self.large_insertion_cutoff:
                            large_insertion_3p = True
                        else:
                            small_insertion_3p = True
                    elif position == block2[0]:
                        insertion_length = pysamtools.get_insertion_length(position, read, reverse=True)
                        if insertion_length >= self.large_insertion_cutoff:
                            large_insertion_5p = True
                        else:
                            small_insertion_5p = True
                else:
                    deletion_range = (block1[1], block2[0])
                    if self.block_overlaps_site(deletion_range, position):
                        deletion = True

        if small_insertion_5p:
            return 'small_insertion_5p'
        elif small_insertion_3p:
            return 'small_insertion_3p'
        elif large_insertion_5p:
            return 'large_insertion_5p'
        elif large_insertion_3p:
            return 'large_insertion_3p'
        elif deletion:
            return 'deletion'
        elif runthrough:
            return 'runthrough'
        else:
            None


    def make_consensus_sequences(self):

        if self.verbose:
            logger.info('Generating consensus sequences from softclipped termini...')
        for contig in self.softclipped_sites:

            for pos in self.softclipped_sites[contig]:

                site = self.softclipped_sites[contig][pos]

                if site.keep_softclips_5p:

                    reads = site.softclip_5p_reads
                    softclip_consensus = SoftclipConsensus(reads, '5p', pos, self.min_count_consensus)
                    site.consensus_sequences_5p = softclip_consensus.consensus_seqs

                if site.keep_softclips_3p:

                    reads = site.softclip_3p_reads
                    softclip_consensus = SoftclipConsensus(reads, '3p', pos, self.min_count_consensus)
                    site.consensus_sequences_3p = softclip_consensus.consensus_seqs


    def block_overlaps_site(self, block, position):

        if block[0] <= position and block[1] > position:
            return True
        else:
            return False

    def has_nearby_5p_mate(self, pos, positions):
        closest_5p = misc.takeClosestLarger(positions, pos)
        if closest_5p is None:
            return False
        if closest_5p - pos <= self.min_distance_to_mate:
            return True
        else:
            return False

    def has_nearby_3p_mate(self, pos, positions):
        closest_3p = misc.takeClosestSmaller(positions, pos)
        if closest_3p is None:
            return False
        if pos - closest_3p <= self.min_distance_to_mate:
            return True
        else:
            return False

    def get_softclip_5p_positions(self):
        positions = defaultdict(list)

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                if self.softclipped_sites[contig][pos].get_softclip_5p_count() > 0:
                    positions[contig].append(pos)

        for contig in positions:
            positions[contig] = sorted(positions[contig])

        return positions

    def get_softclip_3p_positions(self):
        positions = defaultdict(list)

        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                if self.softclipped_sites[contig][pos].get_softclip_3p_count() > 0:
                    positions[contig].append(pos)

        for contig in positions:
            positions[contig] = sorted(positions[contig])

        return positions

    def passes_read_filters(self, read):
        if read.mapping_quality < self.min_alignment_quality:
            return False
        elif not sctools.read_meets_min_alignment_inner_length(read, self.min_alignment_inner_length):
            return False
        else:
            return True

    def parse_right_softclipped_read(self, read):
        meets_minlength = self.meets_minlength_right(read)
        contig, pos = sctools.right_softclipped_site_lenient(read)
        self.softclipped_sites[contig][pos].add_softclip_5p(read, meets_minlength)

    def parse_left_softclipped_read(self, read):
        meets_minlength = self.meets_minlength_left(read)
        contig, pos = sctools.left_softclipped_site_lenient(read)
        self.softclipped_sites[contig][pos].add_softclip_3p(read, meets_minlength)

    def meets_minlength_right(self, read):
        meets_minlength = False
        if sctools.right_softclip_length(read) >= self.min_softclip_length:
            meets_minlength = True
        return meets_minlength

    def meets_minlength_left(self, read):
        meets_minlength = False
        if sctools.left_softclip_length(read) >= self.min_softclip_length:
            meets_minlength = True
        return meets_minlength


    def count_softclips(self):
        count = 0
        for contig in self.softclipped_sites:
            count += len(self.softclipped_sites[contig])
        return count

    def count_consensus_seqs(self):
        count = 0
        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                count += len(self.softclipped_sites[contig][pos].consensus_sequences_5p)
                count += len(self.softclipped_sites[contig][pos].consensus_sequences_3p)
        return count

    def make_dataframe(self):

        column_names = ['contig', 'pos', 'orient', 'softclip_count_5p', 'softclip_count_3p', 'runthrough_count',
                        'small_insertion_count_5p', 'small_insertion_count_3p',
                        'large_insertion_count_5p', 'large_insertion_count_3p', 'deletion_count',
                        'upstream_deletion_count', 'downstream_deletion_count', 'total_count',
                        'consensus_softclip_count', 'consensus_seq']

        outdata= dict()
        for contig in self.softclipped_sites:
            sorted_positions = sorted(list(self.softclipped_sites[contig].keys()))

            for pos in sorted_positions:
                site = self.softclipped_sites[contig][pos]

                if site.keep_softclips_5p:

                    for consensus_sequence in site.consensus_sequences_5p:

                        outdata[len(outdata)] = [
                            contig, pos, '5p', site.get_softclip_5p_count(), site.get_softclip_3p_count(),
                            site.get_runthrough_count(), site.get_small_insertion_5p_count(), site.get_small_insertion_3p_count(),
                            site.get_large_insertion_5p_count(), site.get_large_insertion_3p_count(),
                            site.get_deletion_count(), site.get_upstream_deletion_count(), site.get_downstream_deletion_count(),
                            site.get_total_count(), consensus_sequence[0], consensus_sequence[1]
                        ]

                if site.keep_softclips_3p:
                    for consensus_sequence in site.consensus_sequences_3p:
                        outdata[len(outdata)] = [
                            contig, pos, '3p', site.get_softclip_5p_count(), site.get_softclip_3p_count(),
                            site.get_runthrough_count(), site.get_small_insertion_5p_count(), site.get_small_insertion_3p_count(),
                            site.get_large_insertion_5p_count(), site.get_large_insertion_3p_count(),
                            site.get_deletion_count(), site.get_upstream_deletion_count(), site.get_downstream_deletion_count(),
                            site.get_total_count(), consensus_sequence[0], consensus_sequence[1]
                        ]

        out_df = pd.DataFrame.from_dict(outdata, orient='index', columns=column_names)

        return out_df


    def print_sites(self):

        print('contig\tpos\ttotal\trunthrough_count\tsoftclip_5p_count\tsoftclip_3p_count'
              '\tsmall_insertions_5p_count\tsmall_insertions_3p_count\t'
              'large_insertions_5p_count\tlarge_insertions_3p_count\tdeletion_count', file=sys.stderr)
        for contig in self.softclipped_sites:
            for pos in self.softclipped_sites[contig]:
                site = self.softclipped_sites[contig][pos]
                print('{contig}\t{pos}\t'.format(contig=contig, pos=pos) + str(site), file=sys.stderr)


class SoftclipSite:

    softclip_5p_reads = None
    softclip_3p_reads = None

    meets_minlength_5p = None
    meets_minlength_3p = None

    insertion_5p_reads = None
    insertion_3p_reads = None
    runthrough_reads = None
    deletion_reads = None
    upstream_deletion_reads = None
    downstream_deletion_reads = None

    keep_softclips_5p = None
    keep_softclips_3p = None

    consensus_sequences_5p = None
    consensus_sequences_3p = None

    def __init__(self):
        self.softclip_5p_reads = defaultdict(set)
        self.softclip_3p_reads = defaultdict(set)

        self.meets_minlength_5p = False
        self.meets_minlength_3p = False

        self.small_insertion_5p_reads = set()
        self.small_insertion_3p_reads = set()
        self.large_insertion_5p_reads = set()
        self.large_insertion_3p_reads = set()
        self.runthrough_reads = set()
        self.deletion_reads = set()
        self.upstream_deletion_reads = set()
        self.downstream_deletion_reads = set()

        self.keep_softclips_5p = True
        self.keep_softclips_3p = True

        self.consensus_sequences_5p = []
        self.consensus_sequences_3p = []

    def add_softclip_5p(self, read, meets_minlength):
        self.softclip_5p_reads[read.query_name].add(read)

        if not self.meets_minlength_5p and meets_minlength:
            self.meets_minlength_5p = True

    def add_softclip_5p_reads(self, reads):
        for read in reads:
            self.softclip_5p_reads[read.query_name].add(read)

    def add_softclip_3p(self, read, meets_minlength):
        self.softclip_3p_reads[read.query_name].add(read)

        if not self.meets_minlength_3p and meets_minlength:
            self.meets_minlength_3p = True

    def add_softclip_3p_reads(self, reads):
        for read in reads:
            self.softclip_3p_reads[read.query_name].add(read)

    def add_runthrough(self, read):
        self.runthrough_reads.add(read.query_name)

    def add_runthrough_reads(self, reads):
        self.runthrough_reads.update(reads)

    def add_small_insertion_5p(self, read):
        self.small_insertion_5p_reads.add(read.query_name)

    def add_small_insertion_5p_reads(self, reads):
        self.small_insertion_5p_reads.update(reads)

    def add_large_insertion_5p(self, read):
        self.large_insertion_5p_reads.add(read.query_name)

    def add_large_insertion_5p_reads(self, reads):
        self.large_insertion_5p_reads.update(reads)

    def add_small_insertion_3p(self, read):
        self.small_insertion_3p_reads.add(read.query_name)

    def add_small_insertion_3p_reads(self, reads):
        self.small_insertion_3p_reads.update(reads)

    def add_large_insertion_3p(self, read):
        self.large_nsertion_3p_reads.add(read.query_name)

    def add_large_insertion_3p_reads(self, reads):
        self.large_insertion_3p_reads.update(reads)

    def add_deletion(self, read):
        self.deletion_reads.add(read.query_name)

    def add_deletion_reads(self, reads):
        self.deletion_reads.update(reads)

    def add_upstream_deletion_reads(self, reads):
        self.upstream_deletion_reads.update(reads)

    def add_downstream_deletion_reads(self, reads):
        self.downstream_deletion_reads.update(reads)

    def get_total_count(self):

        total_count = self.get_softclip_5p_count() + self.get_softclip_3p_count() + self.get_runthrough_count() + \
            self.get_small_insertion_5p_count()   + self.get_small_insertion_3p_count() + \
            self.get_large_insertion_5p_count() + self.get_large_insertion_3p_count() + self.get_deletion_count()

        return total_count

    def get_softclip_5p_count(self):
        return len(self.softclip_5p_reads)

    def get_softclip_3p_count(self):
        return len(self.softclip_3p_reads)

    def get_runthrough_count(self):
        return len(self.runthrough_reads)

    def get_small_insertion_5p_count(self):
        return len(self.small_insertion_5p_reads)

    def get_small_insertion_3p_count(self):
        return len(self.small_insertion_3p_reads)

    def get_large_insertion_5p_count(self):
        return len(self.large_insertion_5p_reads)

    def get_large_insertion_3p_count(self):
        return len(self.large_insertion_3p_reads)

    def get_deletion_count(self):
        return len(self.deletion_reads)

    def get_upstream_deletion_count(self):
        return len(self.upstream_deletion_reads)

    def get_downstream_deletion_count(self):
        return len(self.downstream_deletion_reads)

    def get_softclip_ratio_5p(self):
        return (self.get_softclip_5p_count() + self.get_large_insertion_5p_count() +
                self.get_large_insertion_3p_count()) / self.get_total_count()

    def get_softclip_ratio_3p(self):
        return (self.get_softclip_3p_count() + self.get_large_insertion_5p_count() +
                self.get_large_insertion_3p_count()) / self.get_total_count()

    def get_upstream_deletion_ratio(self):
        return self.get_upstream_deletion_count() / self.get_total_count()

    def get_downstream_deletion_ratio(self):
        return self.get_downstream_deletion_count() / self.get_total_count()

    def get_indel_ratio_5p(self):
        return (self.get_small_insertion_5p_count() + self.get_deletion_count()) / self.get_total_count()

    def get_indel_ratio_3p(self):
        return (self.get_small_insertion_3p_count() + self.get_deletion_count()) / self.get_total_count()

    def __str__(self):
        outstring = '{total}\t{runthrough_count}\t{softclip_5p_count}\t{softclip_3p_count}' \
                    '\t{small_insertions_5p_count}\t{small_insertions_3p_count}\t' \
                    '{large_insertions_5p_count}\t{large_insertions_3p_count}\t{deletion_count}'.format(
            total=self.get_total_count(), runthrough_count=self.get_runthrough_count(),
            softclip_5p_count=self.get_softclip_5p_count(), softclip_3p_count=self.get_softclip_3p_count(),
            small_insertions_5p_count=self.get_small_insertion_5p_count(),
            small_insertions_3p_count=self.get_small_insertion_3p_count(),
            large_insertions_5p_count=self.get_large_insertion_5p_count(),
            large_insertions_3p_count=self.get_large_insertion_3p_count(),
            deletion_count=self.get_deletion_count())

        return outstring


class SoftclipConsensus:

    reads = None
    orient = None
    softclip_pos = None
    min_count_consensus = None
    softclipped_seqs = None
    softclipped_qualities = None
    consensus_seqs = None

    def __init__(self, reads, orient, softclip_pos, min_count_consensus):
        self.reads = reads
        self.orient = orient
        self.softclip_pos = softclip_pos
        self.min_count_consensus = min_count_consensus
        self.softclipped_seqs = []
        self.softclipped_qualities = []
        self.consensus_seqs = []

        self.get_softclipped_sequences_and_qualities()
        self.make_consensus()


    def make_consensus(self):

        terminustrie = self.get_terminus_sequence_trie(self.softclipped_seqs, self.softclipped_qualities)
        terminustrie_traversal = terminustrie.traverse_all()
        seq_dict = {seq[0]: [seq[1], seq[2]] for seq in terminustrie_traversal}
        seq_clusters = self.get_sequence_clusters(seq_dict)

        consensus_seqs = self.get_cluster_consensus_seqs(seq_clusters)

        if self.orient == '3p':
            consensus_seqs = [seq[::-1] for seq in consensus_seqs]

        merged_counts = self.merge_cluster_counts(seq_clusters, terminustrie)

        self.consensus_seqs = zip(merged_counts, consensus_seqs)


    def get_terminus_sequence_trie(self, softclipped_seqs, softclipped_base_qual):

        mytrie = terminustrie.Trie()

        for i in range(len(softclipped_seqs)):
            seq, qual = softclipped_seqs[i], softclipped_base_qual[i]
            mytrie.add(seq.upper(), qual)

        return mytrie


    def get_softclipped_sequences_and_qualities(self):

        self.softclipped_seqs = []
        self.softclipped_qualities = []

        for read in self.reads:
            reads = self.reads[read]
            if self.orient == '3p':

                seq, quals = self.combine_softclip_seq_quals_3p(reads)

            elif self.orient == '5p':

                seq, quals = self.combine_softclip_seq_quals_5p(reads)

            self.softclipped_seqs.append(seq)
            self.softclipped_qualities.append(quals)


    def combine_softclip_seq_quals_3p(self, reads):

        if len(reads) == 1:
            read = reads.pop()
            return sctools.left_softclipped_sequence(read)[::-1], sctools.left_softclip_qualities(read)[::-1]

        elif len(reads) == 2:
            read1 = reads.pop()
            read2 = reads.pop()

            read1_qualities = sctools.left_softclip_qualities(read1)[::-1]
            read2_qualities = sctools.left_softclip_qualities(read2)[::-1]

            read1_clipped_seq = sctools.left_softclipped_sequence(read1)[::-1]
            read2_clipped_seq = sctools.left_softclipped_sequence(read2)[::-1]

            outseq = ''
            outquals = []

            for pos in range(max([len(read1_qualities), len(read2_qualities)])):

                if pos < len(read1_qualities) and pos < len(read2_qualities):
                    qual1, qual2 = read1_qualities[pos], read2_qualities[pos]
                    char1, char2 = read1_clipped_seq[pos], read2_clipped_seq[pos]

                    if qual1 >= qual2:
                        outquals.append(qual1)
                        outseq += char1
                    else:
                        outquals.append(qual2)
                        outseq += char2

                elif pos < len(read1_qualities):
                    outquals.append(read1_qualities[pos])
                    outseq += read1_clipped_seq[pos]
                elif pos < len(read2_qualities):
                    outquals.append(read2_qualities[pos])
                    outseq += read2_clipped_seq[pos]
                else:
                    print('ERROR: YOU MISSED SOMETHING')
                    sys.exit()


            return outseq, outquals
        else:
            print("READ NUMBER ERROR - ZERO OR MORE THAN TWO READS ASSOCIATED WITH QUERY NAME")
            sys.exit()


    def combine_softclip_seq_quals_5p(self, reads):

        if len(reads) == 1:
            read = reads.pop()
            return sctools.right_softclipped_sequence(read), sctools.right_softclip_qualities(read)

        elif len(reads) == 2:
            read1 = reads.pop()
            read2 = reads.pop()

            read1_qualities = sctools.right_softclip_qualities(read1)
            read2_qualities = sctools.right_softclip_qualities(read2)

            read1_clipped_seq = sctools.right_softclipped_sequence(read1)
            read2_clipped_seq = sctools.right_softclipped_sequence(read2)

            outseq = ''
            outquals = []

            for pos in range(max([len(read1_qualities), len(read2_qualities)])):

                if pos < len(read1_qualities) and pos < len(read2_qualities):
                    qual1, qual2 = read1_qualities[pos], read2_qualities[pos]
                    char1, char2 = read1_clipped_seq[pos], read2_clipped_seq[pos]

                    if qual1 >= qual2:
                        outquals.append(qual1)
                        outseq += char1
                    else:
                        outquals.append(qual2)
                        outseq += char2

                elif pos < len(read1_qualities):
                    outquals.append(read1_qualities[pos])
                    outseq += read1_clipped_seq[pos]
                elif pos < len(read2_qualities):
                    outquals.append(read2_qualities[pos])
                    outseq += read2_clipped_seq[pos]
                else:
                    print('ERROR: YOU MISSED SOMETHING')
                    sys.exit()


            return outseq, outquals
        else:
            print("READ NUMBER ERROR - ZERO OR MORE THAN TWO READS ASSOCIATED WITH QUERY NAME")
            sys.exit()



    def get_sequence_clusters(self, sequence_dict, perc_similarity=0.75):

        if len(sequence_dict) == 1:
            return [sequence_dict]

        sequences = list(sequence_dict.keys())

        no_match = set(sequences)
        has_match = set()
        pairwise_matches = set()
        for i, j in itertools.combinations(range(len(sequences)), r=2):

            seq1, seq2 = sequences[i], sequences[j]
            minlength = min([len(seq1), len(seq2)])
            dist = levenshtein_distance(seq1[:minlength], seq2[:minlength])
            similarity = 1 - (dist / minlength)

            if similarity >= perc_similarity:
                has_match.add(i)
                has_match.add(j)
                if seq1 in no_match: no_match.remove(seq1)
                if seq2 in no_match: no_match.remove(seq2)
                pairwise_matches.add((i, j))

        pair_order = sorted(list(has_match))
        adjacency_matrix = pd.DataFrame(np.zeros((len(pair_order), len(pair_order))), index=pair_order,
                                        columns=pair_order)

        for pair in pairwise_matches:
            adjacency_matrix.loc[pair[0], pair[1]] = 1
            adjacency_matrix.loc[pair[1], pair[0]] = 1

        n_comp, assignments = connected_components(adjacency_matrix.values)
        final_components = []
        for i in range(n_comp):
            comp = list(np.array(pair_order)[assignments == i])
            final_components.append(comp)

        final_clusters = [{seq: sequence_dict[seq]} for seq in no_match]

        for comp in final_components:
            new_seq_cluster = dict()
            for i in comp:
                new_seq_cluster[sequences[i]] = sequence_dict[sequences[i]]
            final_clusters.append(new_seq_cluster)

        return final_clusters

    def get_cluster_consensus_seqs(self, seq_clusters):
        consensus_seqs = list()

        for clust in seq_clusters:
            seqs = [seq for seq in clust]
            quals = [clust[seq][0] for seq in clust]
            counts = [clust[seq][1] for seq in clust]

            mytrie = terminustrie.Trie()
            mytrie.load_words(seqs, quals, counts)
            consensus = mytrie.make_consensus_word(self.min_count_consensus)

            consensus_seqs.append(consensus)

        return consensus_seqs

    def merge_cluster_counts(self, seq_clusters, terminustrie):

        seq_clusters = [list(cluster.keys()) for cluster in seq_clusters]
        out = []
        for clust in seq_clusters:
            subtrie = terminustrie.make_subtrie(clust)
            total_words = subtrie.total_words
            out.append(total_words)

        return out

    def print_all(self):
        if self.orient == '5p':
            print('SOFTCLIPPED SEQS:')
            for seq in self.softclipped_seqs:
                print('\t'+seq)
            print('CONSENSUS SEQS:')
            for seq in self.consensus_seqs:
                print('\t' + seq[1])
            print()
        else:
            print('SOFTCLIPPED SEQS:')
            for seq in self.softclipped_seqs:
                print('\t' + seq[::-1])
            print('CONSENSUS SEQS:')
            for seq in self.consensus_seqs:
                print('\t' + seq[1])
            print()


if __name__ == '__main__':

    softclip_parser = SoftclipParser(pysam.AlignmentFile(sys.argv[1], 'rb'))
    softclip_parser.parse_softclips()
    softclip_parser.filter_softclips_minlength()
    softclip_parser.filter_softclips_mincount()
    softclip_parser.filter_softclips_mindistance()
    softclip_parser.parse_unclipped_read_info()
    softclip_parser.filter_softclips_count_ratios()
    softclip_parser.filter_softclips_mindistance()
    softclip_parser.make_consensus_sequences()
    softclip_parser.filter_consensus_sequences_minlength()
    softclip_parser.filter_consensus_sequences_mincount()
    softclip_parser.filter_softclips_mindistance()

    print(softclip_parser.make_dataframe().to_string())
